/*========================================================================
 * Copyright (c) 2009 Tampere University of Technology, Finland
 * 
 * Permission is hereby granted, free of charge, to any person 
 * obtaining a copy of this software and associated documentation 
 * files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/

function tr(s) { return s; }

/* Main application widget */
/* This application uses Webcams.Travel API and Google Maps API to show
	webcam images over a Google Map. */
function QWebcamMap(parent) {
	QWidget.call(this, parent);
	
	// Application is disabled until webview is fully loaded
	this.setDisabled(true);
	
	// Google Maps API key
	// Get your own API key from Google: http://code.google.com/intl/fi-FI/apis/maps/signup.html
	this.MY_API_KEY = "ABQIAAAAjfRjTDOGscFuDUwMxJYaQRQUpi_xkJDYa0W0Jatk4awxKuvp7BQBfOyxYXrIsE8mc8pdYrHlaNjZjQ";
	
	// Webcams.traver devid
	this.WT_DEVID = "0ada714bb65ef9c8ff81f4ca329e06dc";
	this.WT_REST_REQ = "http://api.webcams.travel/rest?method=";
	
	// Array to hold the small images over the main map
	this.camArray = new Array();
	
	// QWebView shows a web page with full screen Google Map
	this.view = new QWebView(this);
	
	// lineEdit and searchButton are in the upper row
	this.lineEdit = new QLineEdit(this);
	this.searchButton = new QPushButton(tr("Search"), this);
		
	// Use timer to check if the map center has been changed
	this.centerChangedTimer = new QTimer(this);
	this.centerChangedTimer["timeout"].connect(this, this.centerChangedTimerTimeout);
	this.centerChangedTimer.start(2000);
	
	// Use timer to check if user has clicked a marker or not
	this.cameraTimer = new QTimer(this);
	this.cameraTimer["timeout"].connect(this, this.timerTimeout);

	// upper row widget connections
	this.searchButton["clicked"].connect(this, this.searchAddress);
	this.lineEdit["returnPressed"].connect(this, this.searchAddress);

	// let's do some initialization after view has loaded its web page
	this.view["loadFinished(bool)"].connect(this, this.loadFinished);

	// Complex layout definitions
	this.layout = new QVBoxLayout(this);
	this.hboxLayout = new QHBoxLayout();
	
	// Add widgets to the layout
	this.hboxLayout.addWidget(this.lineEdit, 0, 0);
	this.hboxLayout.addWidget(this.searchButton, 0, 0);
		
	this.layout.addLayout(this.hboxLayout);
	this.layout.addWidget(this.view, 1, 0);
	//this.layout.addLayout(this.smallImageLayout);

	
	// Try to load a web page with a Google Map
	try {
		this.view.load(new QUrl("http://www.cs.tut.fi/~salmin39/map/index.html"));
	} catch(err) {
		var messageBox = new QMessageBox(this);
		messageBox.text = "Error: " + err;
		messageBox.show();
	}
	
	this.windowTitle = tr("QWebcamMap");
	this.resize(800,480);
	
	// Nice effect
	this.darkFader = new FaderWidget(this,0);
	this.darkFader.start();
}

QWebcamMap.prototype = new QWidget();

/* Set map center and zoom */
QWebcamMap.prototype.setCenter = function(lat,lng,zoom) {
	var scriptString = 
	"map.setCenter(new GLatLng(" + lat + "," + lng + ")," + zoom + ");";
	this.view.page().mainFrame().evaluateJavaScript(scriptString);
}

// Check if the map center has been changed and
// if it is, get new cams
QWebcamMap.prototype.centerChangedTimerTimeout = function() {
	if(this.mapCenter.x.toString() != this.getMapCenter().x.toString() ||
		this.mapCenter.y.toString() != this.getMapCenter().y.toString()) {
		this.getCams();
		this.mapCenter = this.getMapCenter();
	}
}

/* If user has clicked a marker we should show marker related weather
 * information if it is available */
QWebcamMap.prototype.timerTimeout = function() {
	// get the marker
	var marker = this.getUserLastClickedMarker();
	this.view.page().mainFrame().evaluateJavaScript("userClickedMarker = undefined;");
	if(marker != undefined && marker.camID != undefined) {
		try{
			var webcam = new Object()
			webcam.latitude = marker.ra.y; // Critical place when Google Maps changes its marker definitions
			webcam.longitude = marker.ra.x;
			webcam.webcamid = marker.camID;
			
			var cam = new ZoomImageWidget(this, webcam);
		
			if(Math.random() > 0.5)
				xPos = Math.floor((this.width/2-250)*Math.random()+10);
			else
				xPos = Math.floor((this.width/2-250)*Math.random()+this.width/2+50);

			cam.pos = new QPoint(xPos,
				Math.floor((this.height-300)*Math.random()+50));	
			cam.show();
			this.camArray.push(cam);
		}
		catch(err) {
			print(err);
		}
	}
}

/* After the web page has been fully loaded we can do some
 * initialization and other stuff. If this is done too soon it will
 * not work at all */
QWebcamMap.prototype.loadFinished = function(ok) {	
	var scriptString = 
	"var userClickedPoint = undefined;" +
	"var userClickedMarker = undefined;" +
	"var myClickListener = undefined;" +
	//"map.addControl(new GMapTypeControl());" +
	"map.setCenter( new GLatLng(0,0),1 );" +
	"var markers = new Array();" +
	//"var mm = new GMarkerManager(map, {borderPadding:1});" +
	"map.disableDoubleClickZoom();" +
	"var topRight = new GControlPosition(G_ANCHOR_TOP_RIGHT, new GSize(10,10));"+
	"map.addControl(new GLargeMapControl(), topRight);" +
	"myClickListener = GEvent.addListener(map, \"click\", function(marker, point) {" +
    	"userClickedPoint = point;" +
    	"userClickedMarker = marker;" +
	"});" +
	"GEvent.addListener(map, \"moveend\", function() {" +
  		"userClickedPoint = undefined;" +
  		"userClickedMarker = undefined;" +
	"});" +

	// Customizes lettered icons
	'var baseIcon = new GIcon(G_DEFAULT_ICON);' +
	'baseIcon.shadow = \"http://www.google.com/mapfiles/shadow50.png\";'+
	'baseIcon.iconSize = new GSize(20, 34);'+
	'baseIcon.shadowSize = new GSize(37, 34);'+
	'baseIcon.iconAnchor = new GPoint(9, 34);'+
	'baseIcon.infoWindowAnchor = new GPoint(9, 2);'+

	'function createMarker(point, index, infoHtml, camID) {'+
		'var letter = String.fromCharCode(\"A\".charCodeAt(0) + index);'+
		'var letteredIcon = new GIcon(baseIcon);'+
		'letteredIcon.image = \"http://www.google.com/mapfiles/marker\" + letter + \".png\";'+

		'markerOptions = { icon:letteredIcon, title:infoHtml.toString() };'+
		'var marker = new GMarker(point, markerOptions);'+
		'marker.letter = letter;'+
		'marker.camID = camID;' +
		'return marker;'+
	'}';
	
	this.view.page().mainFrame().evaluateJavaScript(scriptString);
	
	this.mapCenter = this.getMapCenter();
	
	// Enable whole application
	this.setEnabled(true);
}

QWebcamMap.prototype.getMapCenter = function() {
	return this.view.page().mainFrame().evaluateJavaScript("map.getCenter()");
}

QWebcamMap.prototype.getUserLastClickedMarker = function() {
	return this.view.page().mainFrame().evaluateJavaScript("userClickedMarker;");
}

QWebcamMap.prototype.getCams = function() {
	var mapCenter = this.getMapCenter();
	
	// Get camera data from Webcams Travel (http://fi.webcams.travel/)
	requestStr = this.WT_REST_REQ + "wct.webcams.list_nearby&devid="
		+ this.WT_DEVID
		+ "&lat=" + mapCenter.y.toString()
		+ "&lng=" + mapCenter.x.toString()
		+ "&radius=10&unit=km&per_page=50&format=json";
		
	var accessManager = new QNetworkAccessManager(this);
	accessManager["finished(QNetworkReply*)"].connect(this, this.getCamsCallback);
	accessManager.get(new QNetworkRequest(new QUrl(requestStr)));
}

QWebcamMap.prototype.getCamsCallback = function(reply) {
	var replyString = reply.readAll().toString();
	var replyJSON = undefined;
	replyJSON = eval('(' + replyString + ')');
	if(replyJSON.status != "ok") {
		var messageBox = new QMessageBox(this);
		messageBox.text = "Webcams.travel error!\n"
			+ "Status: " + replyJSO.status + "\n" 
			+ "Description: " + replyJSON.error.description;
		messageBox.show();
	}
	else if (replyJSON.webcams.count > 0) {
		// Close old open camera images
		while(this.camArray.length > 0)
			this.camArray.pop().close();
		
		// Clear old markers
		this.view.page().mainFrame().evaluateJavaScript("map.clearOverlays();");
		
		// Only 15 cameras at a time
		if(replyJSON.webcams.count > 15)
			replyJSON.webcams.count = 15;
			
		// For each camera: insert marker to the main map
		// and show webcam image on separate widget over the main map
		for(var i = 0; i < replyJSON.webcams.count; i=i+1) {
			this.insertMarker(replyJSON.webcams.webcam[i].latitude,
				replyJSON.webcams.webcam[i].longitude, 
				replyJSON.webcams.webcam[i].title,
				replyJSON.webcams.webcam[i].webcamid, i);
			
			var cam = new ZoomImageWidget(this, replyJSON.webcams.webcam[i]);
			
			// Position the widget on the main map randomly 
			// like in this clarifying picture here: [X| |X]
			// Right or left side of the map
			if(Math.random() > 0.5)
				xPos = Math.floor((this.width/2-250)*Math.random()+10);
			else
				xPos = Math.floor((this.width/2-250)*Math.random()+this.width/2+50);
	
			cam.pos = new QPoint(xPos,
				Math.floor((this.height-300)*Math.random()+50));	
			cam.show();
			this.camArray.push(cam);
		}
		
		// Start a timer to check if the user has clicked a marker
		this.cameraTimer.start(500);
	}
}

QWebcamMap.prototype.geoCode = function(locationString, callback) {
	requestStr = "http://maps.google.com/maps/geo?q=" +
		locationString + "&output=csv&key=" + this.MY_API_KEY;
	var accessManager = new QNetworkAccessManager(this);
	accessManager["finished(QNetworkReply*)"].connect(this, callback);
	accessManager.get(new QNetworkRequest(new QUrl(requestStr)));
}

/* Insert new marker to the map */
QWebcamMap.prototype.insertMarker = function(lat,lng,infoHtml,camID,i) {
	var scriptString =
	'var newMarker = createMarker(new GLatLng(' + lat + ',' + lng + '),' + i + ',\"' + infoHtml + '\",'+ camID +');'+
	'markers.push(newMarker);' +
	'map.addOverlay(markers[markers.length-1]);';
	this.view.page().mainFrame().evaluateJavaScript(scriptString);
}

/* Geocoding an address */
QWebcamMap.prototype.searchAddress = function() {
	this.geoCode(this.lineEdit.text, searchAddressCallback);
}

searchAddressCallback = function(reply) {
	var replyString = reply.readAll().toString();
	var geoCodeArray = new Array();
	geoCodeArray = replyString.split(",");
	var htmlResponse = geoCodeArray[0];
	var accuracy = geoCodeArray[1];
	var lat = geoCodeArray[2];
	var lng = geoCodeArray[3];
	if(htmlResponse == "200") {
		this.setCenter(lat, lng, parseInt(accuracy) + 7);
	}
}

/* After image is succesfully loaded, scale it and show it */
QWebcamMap.prototype.addSmallImageReplyFinished = function(reply) {
	var data = reply.readAll();
	var imagePixmap = new QPixmap();
	var newLabel = new QLabel(this);
	// mouseDoubleClickEvent
	imagePixmap.loadFromData(data, 0, Qt.AutoColor);
	if(!imagePixmap.isNull()) {
		imagePixmap = imagePixmap.scaledToWidth(40,Qt.SmoothTransformation);
		newLabel.setPixmap(imagePixmap);
		~imagePixmap;
	}
	else {
		/* imagePixmap was null, image data was not available (bad url)*/
		newLabel.text = '<font color="white"><center>No image available</center></font>';
	}
	this.smallImageLayout.addWidget(newLabel, 0, Qt.AlignCenter);
}

// A Widget that will show an image and zoom it
// A small map can be viewed over the zoomed in image.
function ZoomImageWidget(parent, webcam) {	
	QWidget.call(this,parent);
	
	this.styleSheet = "QWidget { background-color: rgba(49,100,199,55%); } ";
	
	this.MY_API_KEY = "ABQIAAAAjfRjTDOGscFuDUwMxJYaQRQUpi_xkJDYa0W0Jatk4awxKuvp7BQBfOyxYXrIsE8mc8pdYrHlaNjZjQ";
	this.RESIZE_FACTOR = 4;
	
	this.img = new QPixmap();
	this.map = new QPixmap();
	
	this.setAttribute(Qt.WA_DeleteOnClose);
	this.cursor = new QCursor(Qt.PointingHandCursor);
	
	this.parent = parent;
	this.webcam = webcam;
	
	this.loadImage('http://images.webcams.travel/webcam/' + webcam.webcamid + '.jpg');
}

ZoomImageWidget.prototype = new QWidget();


ZoomImageWidget.prototype.loadMap = function(url) {
	var accessManager = new QNetworkAccessManager(this);
	accessManager["finished(QNetworkReply*)"].connect(this, this.mapReplyFinished);
	accessManager.get(new QNetworkRequest((new QUrl(url))));
}

ZoomImageWidget.prototype.mapReplyFinished = function(reply) {
	var data = reply.readAll();
	this.map.loadFromData(data, Qt.AutoColor);
	this.update();
}

// Zoom the image and load a small map over the image
ZoomImageWidget.prototype.enterEvent = function(event) {
	this.raise();
	this.resize(this.width * this.RESIZE_FACTOR,this.height * this.RESIZE_FACTOR);
	this.loadMap('http://maps.google.com/staticmap?center='
		+this.webcam.latitude+','+this.webcam.longitude+
		'&zoom=13&size=100x100&maptype=mobile&'+'markers='
		+this.webcam.latitude+','+this.webcam.longitude
		+'&sensor=false&key=' + this.MY_API_KEY);
}

// UnZoom the image
ZoomImageWidget.prototype.leaveEvent = function(event) {
	this.resize(this.width/this.RESIZE_FACTOR,this.height/this.RESIZE_FACTOR);
	this.map = new QPixmap();
	this.update();
}

/* Load image frow the web */
ZoomImageWidget.prototype.loadImage = function(url) {
	var accessManager = new QNetworkAccessManager(this);
	accessManager["finished(QNetworkReply*)"].connect(this, this.imageReplyFinished);
	accessManager.get(new QNetworkRequest((new QUrl(url))));
}

/* After the image is succesfully loaded scale the widget */
ZoomImageWidget.prototype.imageReplyFinished = function(reply) {
	var data = reply.readAll();
	this.img.loadFromData(data, Qt.AutoColor);
	
	//this.resize(this.width/2,this.height/2);
	this.resize((this.img.size().width())/this.RESIZE_FACTOR,(this.img.size().height())/this.RESIZE_FACTOR);
}

/* When the widget is dragged with mouse, change widget position */
ZoomImageWidget.prototype.mouseMoveEvent = function(event) {
	var fGeom = this.parent.frameGeometry;
	var offsetY =  this.parent.frameGeometry.y() - this.parent.geometry.y();
	this.pos = new QPoint(event.globalX() - fGeom.x() - this.width/2, event.globalY() - fGeom.y() - this.height/2 + offsetY);
}

// When the user double clicks the widget, the widget is hidden
ZoomImageWidget.prototype.mouseDoubleClickEvent = function(event) {
	this.hide();
}

// When the widget is resized paint the pictures again
// TODO: map is not always cleared if the used makes fast movements with the mouse
ZoomImageWidget.prototype.paintEvent = function (event) {
	var painter = new QPainter();
	painter.begin(this);
	if(!this.img.isNull()) {
		painter.drawPixmap(3,3,this.img.scaledToWidth(this.width-6,Qt.SmoothTransformation));	
	}
	painter.drawPixmap(this.width-this.map.size().width()-6,this.height-this.map.size().height()-6,this.map);	

	painter.end();
}

/* Fadein effect with a widget that destroys itself after a while */
function FaderWidget(parent, endAlpha) {
	QWidget.call(this, parent);
	
    this.startColor = Qt.black;
	this.currentAlpha = 255;
	if(endAlpha)
		this.endAlpha = endAlpha;
	else
		this.endAlpha = 0;
		
	// How long the effect will be
	this.duration = 500; 
	
	this.timer = new QTimer(this);
	this.timer["timeout"].connect(this, this.update);
	
	// Delete widget when closed
	this.setAttribute(Qt.WA_DeleteOnClose);
	
	// Fill the full screen
	this.resize(parent.width, parent.height);
}

FaderWidget.prototype = new QWidget();

// Start the effect
FaderWidget.prototype.start = function() {
	this.currentAlpha = 255;
    this.timer.start(33);
    this.show();
}

// Doing the fadeout
FaderWidget.prototype.paintEvent = function(event) {
	var painter = new QPainter();
	painter.begin(this);
	var semiTransparentColor = new QColor(this.startColor);
	semiTransparentColor.setAlpha(this.currentAlpha);
	painter.fillRect(this.rect, semiTransparentColor);
	this.currentAlpha -= 255 * this.timer.interval / this.duration;
	//print(this.currentAlpha);
	painter.end();
	if (this.currentAlpha <= this.endAlpha) {
		this.timer.stop();
		this.close();
	}
}

QByteArray.prototype.toString = function()
{
   ts = new QTextStream( this, QIODevice.ReadOnly );
   return ts.readAll();
}

/* To use the application in standalone mode, remove the comment around this code
var map = new QWebcamMap();
map.show();

QCoreApplication.exec();
*/

