/*========================================================================
 * Original copyright (c) 2008-2009 Sun Microsystems Laboratories
 * Prototype library Copyright (c) 2005-2008 Sam Stephenson
 *
 * Additional copyrights for portions of code:
 * Copyright (c) 2008-2009 Tampere University of Technology
 * Copyright (c) 2009 Sesca Mobile Platforms  
 * 
 * Permission is hereby granted, free of charge, to any person 
 * obtaining a copy of this software and associated documentation 
 * files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/

/*========================================================================
 * This is an implementation of the Lively JavaScript application 
 * execution and development environment for the Qt platform.
 * 
 * Quick summary of unimplemented features (as of May 6, 2009):
 * - support for polygon, polyline and simple text object types
 * - drop shadows and multiple selection support
 * - object docking and "snap to grid" capabilities
 * - hover events (mouseover/mouseout support)
 * - scheduler for managing QTimers attached to worlds
 * - gradient colors cannot be created from the user interface yet 
 * 
 * Known bugs:
 * - mouse coordinates are read too precisely for pen-based devices 
 * - QtInspector works only in the Global context
 *
 * Qt-related bugs and issues: 
 * - An unhandled exception may occur when exiting the system
 * - QWidget focus features (e.g., foreground/background
 *   colors/highlights) do not work correctly  
 * - QWidget Z order is not always managed properly (rendering bugs in Qt)
 * - many features (e.g., text editing in the code evaluator) do not 
 *   work correctly when running the system inside the Qt browser plug-in 
 *======================================================================*/

// Note: QtScript supports JavaScript 1.5 only.
// The following three Array methods are specific
// to JavaScript 1.6 and must be added manually.

// NOTE: String.substr function is missing from Qt versions earlier than 4.5.
// We have replaced all uses of 'substr' with 'substring'

// From https://developer.mozilla.org/en/Core_JavaScript_1.5_Reference/Objects/Array/forEach
if (!Array.prototype.forEach) {
  // print("Array.forEach() support added");

  Array.prototype.forEach = function(fun /*, thisp*/)
  {
    var len = this.length;
    if (typeof fun != "function")
      throw new TypeError();

    var thisp = arguments[1];
    for (var i = 0; i < len; i++)
    {
      if (i in this)
        fun.call(thisp, this[i], i, this);
    }
  };
}

if (!Array.prototype.map) {
  // print("Array.map() support added");

  Array.prototype.map = function(fun /*, thisp*/)
  {
    var len = this.length;
    if (typeof fun != "function")
      throw new TypeError();

    var res = new Array(len);
    var thisp = arguments[1];
    for (var i = 0; i < len; i++)
    {
      if (i in this)
        res[i] = fun.call(thisp, this[i], i, this);
    }

    return res;
  };
}

if (!Array.prototype.filter) {
  // print("Array.filter() support added");

  Array.prototype.filter = function(fun /*, thisp*/)
  {
    var len = this.length;
    if (typeof fun != "function")
      throw new TypeError();

    var res = new Array();
    var thisp = arguments[1];
    for (var i = 0; i < len; i++)
    {
      if (i in this)
      {
        var val = this[i]; // in case fun mutates this
        if (fun.call(thisp, val, i, this))
          res.push(val);
      }
    }

    return res;
  };
}

// Qt versions earlier than 4.5 do not support String.substr
if (!String.prototype.substr) {
  // print("String.substr() support added");

  String.prototype.substr = function(start, length) {
    if (!length) return this.substring(start);
    else return this.substring(start, start + length - 1);
  } 
}

/*---------------------------------------------------------------------------
 *  Prototype JavaScript framework, version 1.6.0_rc1
 *  (c) 2005-2008 Sam Stephenson
 *  Prototype is freely distributable under the terms of an MIT-style license.
 *  For details, see the Prototype web site: http://www.prototypejs.org/
 *--------------------------------------------------------------------------*/
// Note that this version is heavily trimmed from its original form!

var Prototype = {
  Version: '1.6.0_rc1',
  Browser: 'Qt',
  K: function(x) { return x }
};

Object.extend = function(destination, source) {

  // TJM: For some reason Qt 4.4 fails if source or destination is null.
  if (!destination) destination = [];
  if (!source) return destination;
  
  for (var property in source)
    destination[property] = source[property];
  return destination;
};

Object.extend(Object, {
  inspect: function(object) {
    try {
      if (object === undefined) return 'undefined';
      if (object === null) return 'null';
      return object.inspect ? object.inspect() : object.toString();
    } catch (e) {
      if (e instanceof RangeError) return '...';
      throw e;
    }
  },

  toQueryString: function(object) {
    return $H(object).toQueryString();
  },

  keys: function(object) {
    var keys = [];
    for (var property in object)
      keys.push(property);
    return keys;
  },

  values: function(object) {
    var values = [];
    for (var property in object)
      values.push(object[property]);
    return values;
  },

  clone: function(object) {
    return Object.extend({ }, object);
  },

  isElement: function(object) {
    return object && object.nodeType == 1;
  },

  isArray: function(object) {
    return object && object.constructor === Array;
  },

  // AT: Added Jan 23, 2009
  isHash: function(object) {
    return object instanceof Hash;
  },

  isFunction: function(object) {
    return typeof object == "function";
  },

  isString: function(object) {
    return typeof object == "string";
  },

  isNumber: function(object) {
    return typeof object == "number";
  },

  isUndefined: function(object) {
    return typeof object == "undefined";
  }
});

Object.extend(Function.prototype, {

  /* AT: QtScript's RegExp implementation cannot handle this
  argumentNames: function() {
    var names = this.toString().match(/^[\s\(]*function[^(]*\((.*?)\)/)[1].split(",").invoke("strip");
    return names.length == 1 && !names[0] ? [] : names;
  },
  */

// AT: Qt-specific code: QtScript does not support RegExp fully
// 'argumentNames' above does not work; the code below is a substitute
// to give us the necessary functionality
  firstArgumentName: function() {
    var str = this.toString();
    var firstParen = str.indexOf('(');
    var lastParen = str.indexOf(')');
    if (firstParen != -1 && lastParen != -1) {
        var substr = str.substring(firstParen+1, lastParen);
        var firstArgName = substr.split(",")[0].strip();
        return firstArgName;
    } else return "";
  },

  bind: function() {
    if (arguments.length < 2 && arguments[0] === undefined) return this;
    var __method = this, args = $A(arguments), object = args.shift();
    return function() {
      return __method.apply(object, args.concat($A(arguments)));
    }
  },

  curry: function() {
    if (!arguments.length) return this;
    var __method = this, args = $A(arguments);
    return function() {
      return __method.apply(this, args.concat($A(arguments)));
    }
  },

  delay: function() {
    /* AT: Qt-specific code removal: setTimeout is not supported
    var __method = this, args = $A(arguments), timeout = args.shift() * 1000;
    return window.setTimeout(function() {
      return __method.apply(__method, args);
    }, timeout);
    */
  },

  wrap: function(wrapper) {
    var __method = this;
    return function() {
      return wrapper.apply(this, [__method.bind(this)].concat($A(arguments)));
    }
  }

});

RegExp.prototype.match = RegExp.prototype.test;

/*--------------------------------------------------------------------------*/

Object.extend(String.prototype, {

  truncate: function(length, truncation) {
    length = length || 30;
    truncation = truncation === undefined ? '...' : truncation;
    return this.length > length ?
      this.slice(0, length - truncation.length) + truncation : String(this);
  },

  strip: function() {
    return this.replace(/^\s+/, '').replace(/\s+$/, '');
  },

  toArray: function() {
    return this.split('');
  },

  succ: function() {
    return this.slice(0, this.length - 1) +
      String.fromCharCode(this.charCodeAt(this.length - 1) + 1);
  },

  times: function(count) {
    return count < 1 ? '' : new Array(count + 1).join(this);
  },

  camelize: function() {
    var parts = this.split('-'), len = parts.length;
    if (len == 1) return parts[0];

    var camelized = this.charAt(0) == '-'
      ? parts[0].charAt(0).toUpperCase() + parts[0].substring(1)
      : parts[0];

    for (var i = 1; i < len; i++)
      camelized += parts[i].charAt(0).toUpperCase() + parts[i].substring(1);

    return camelized;
  },

  capitalize: function() {
    return this.charAt(0).toUpperCase() + this.substring(1).toLowerCase();
  },

  include: function(pattern) {
    return this.indexOf(pattern) > -1;
  },

  startsWith: function(pattern) {
    return this.indexOf(pattern) === 0;
  },

  endsWith: function(pattern) {
    var d = this.length - pattern.length;
    return d >= 0 && this.lastIndexOf(pattern) === d;
  },

  empty: function() {
    return this == '';
  },

  blank: function() {
    return /^\s*$/.test(this);
  }

});

var $break = { };

var Enumerable = {
  each: function(iterator, context) {
    var index = 0;
    iterator = iterator.bind(context);
    try {
      this._each(function(value) {
        iterator(value, index++);
      });
    } catch (e) {
      if (e != $break) throw e;
    }
    return this;
  },

  eachSlice: function(number, iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var index = -number, slices = [], array = this.toArray();
    while ((index += number) < array.length)
      slices.push(array.slice(index, index+number));
    return slices.collect(iterator, context);
  },

  all: function(iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var result = true;
    this.each(function(value, index) {
      result = result && !!iterator(value, index);
      if (!result) throw $break;
    });
    return result;
  },

  any: function(iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var result = false;
    this.each(function(value, index) {
      if (result = !!iterator(value, index))
        throw $break;
    });
    return result;
  },

  collect: function(iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var results = [];
    this.each(function(value, index) {
      results.push(iterator(value, index));
    });
    return results;
  },

  detect: function(iterator, context) {
    iterator = iterator.bind(context);
    var result;
    this.each(function(value, index) {
      if (iterator(value, index)) {
        result = value;
        throw $break;
      }
    });
    return result;
  },

  findAll: function(iterator, context) {
    iterator = iterator.bind(context);
    var results = [];
    this.each(function(value, index) {
      if (iterator(value, index))
        results.push(value);
    });
    return results;
  },

  grep: function(filter, iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var results = [];

    if (Object.isString(filter))
      filter = new RegExp(filter);

    this.each(function(value, index) {
      if (filter.match(value))
        results.push(iterator(value, index));
    });
    return results;
  },

  include: function(object) {
    if (Object.isFunction(this.indexOf))
      if (this.indexOf(object) != -1) return true;

    var found = false;
    this.each(function(value) {
      if (value == object) {
        found = true;
        throw $break;
      }
    });
    return found;
  },

  inGroupsOf: function(number, fillWith) {
    fillWith = fillWith === undefined ? null : fillWith;
    return this.eachSlice(number, function(slice) {
      while(slice.length < number) slice.push(fillWith);
      return slice;
    });
  },

  inject: function(memo, iterator, context) {
    iterator = iterator.bind(context);
    this.each(function(value, index) {
      memo = iterator(memo, value, index);
    });
    return memo;
  },

  invoke: function(method) {
    var args = $A(arguments).slice(1);
    return this.map(function(value) {
      return value[method].apply(value, args);
    });
  },

  max: function(iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var result;
    this.each(function(value, index) {
      value = iterator(value, index);
      if (result == undefined || value >= result)
        result = value;
    });
    return result;
  },

  min: function(iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var result;
    this.each(function(value, index) {
      value = iterator(value, index);
      if (result == undefined || value < result)
        result = value;
    });
    return result;
  },

  partition: function(iterator, context) {
    iterator = iterator ? iterator.bind(context) : Prototype.K;
    var trues = [], falses = [];
    this.each(function(value, index) {
      (iterator(value, index) ?
        trues : falses).push(value);
    });
    return [trues, falses];
  },

  pluck: function(property) {
    var results = [];
    this.each(function(value) {
      results.push(value[property]);
    });
    return results;
  },

  reject: function(iterator, context) {
    iterator = iterator.bind(context);
    var results = [];
    this.each(function(value, index) {
      if (!iterator(value, index))
        results.push(value);
    });
    return results;
  },

  sortBy: function(iterator, context) {
    iterator = iterator.bind(context);
    return this.map(function(value, index) {
      return {value: value, criteria: iterator(value, index)};
    }).sort(function(left, right) {
      var a = left.criteria, b = right.criteria;
      return a < b ? -1 : a > b ? 1 : 0;
    }).pluck('value');
  },

  toArray: function() {
    return this.map();
  },

  zip: function() {
    var iterator = Prototype.K, args = $A(arguments);
    if (Object.isFunction(args.last()))
      iterator = args.pop();

    var collections = [this].concat(args).map($A);
    return this.map(function(value, index) {
      return iterator(collections.pluck(index));
    });
  },

  size: function() {
    return this.toArray().length;
  },

  inspect: function() {
    return '#<Enumerable:' + this.toArray().inspect() + '>';
  }
};

Object.extend(Enumerable, {
  map:     Enumerable.collect,
  find:    Enumerable.detect,
  select:  Enumerable.findAll,
  filter:  Enumerable.findAll,
  member:  Enumerable.include,
  entries: Enumerable.toArray,
  every:   Enumerable.all,
  some:    Enumerable.any
});
function $A(iterable) {
  if (!iterable) return [];
  if (iterable.toArray) return iterable.toArray();
  var length = iterable.length, results = new Array(length);
  while (length--) results[length] = iterable[length];
  return results;
}

Array.from = $A;

Object.extend(Array.prototype, Enumerable);

if (!Array.prototype._reverse) Array.prototype._reverse = Array.prototype.reverse;

Object.extend(Array.prototype, {
  _each: function(iterator) {
    for (var i = 0, length = this.length; i < length; i++)
      iterator(this[i]);
  },

  clear: function() {
    this.length = 0;
    return this;
  },

  first: function() {
    return this[0];
  },

  last: function() {
    return this[this.length - 1];
  },

  compact: function() {
    return this.select(function(value) {
      return value != null;
    });
  },

  flatten: function() {
    return this.inject([], function(array, value) {
      return array.concat(Object.isArray(value) ?
        value.flatten() : [value]);
    });
  },

  without: function() {
    var values = $A(arguments);
    return this.select(function(value) {
      return !values.include(value);
    });
  },

  reverse: function(inline) {
    return (inline !== false ? this : this.toArray())._reverse();
  },

  reduce: function() {
    return this.length > 1 ? this : this[0];
  },

  uniq: function(sorted) {
    return this.inject([], function(array, value, index) {
      if (0 == index || (sorted ? array.last() != value : !array.include(value)))
        array.push(value);
      return array;
    });
  },

  intersect: function(array) {
    return this.uniq().findAll(function(item) {
      return array.detect(function(value) { return item === value });
    });
  },

  clone: function() {
    return [].concat(this);
  },

  size: function() {
    return this.length;
  },

  inspect: function() {
    return '[' + this.map(Object.inspect).join(', ') + ']';
  }

});

// use native browser JS 1.6 implementation if available
if (Object.isFunction(Array.prototype.forEach))
  Array.prototype._each = Array.prototype.forEach;

if (!Array.prototype.indexOf) Array.prototype.indexOf = function(item, i) {
  i || (i = 0);
  var length = this.length;
  if (i < 0) i = length + i;
  for (; i < length; i++)
    if (this[i] === item) return i;
  return -1;
};

if (!Array.prototype.lastIndexOf) Array.prototype.lastIndexOf = function(item, i) {
  i = isNaN(i) ? this.length : (i < 0 ? this.length + i : i) + 1;
  var n = this.slice(0, i).reverse().indexOf(item);
  return (n < 0) ? n : i - n - 1;
};

Array.prototype.toArray = Array.prototype.clone;

// ===========================================================================
// Our extensions to JavaScript base libraries
// ===========================================================================

var Class = {

  // AT: Added from Prototype 1.6.0.2 (www.prototypejs.org)
  create: function() {
    var parent = null, properties = $A(arguments);
    if (Object.isFunction(properties[0]))
      parent = properties.shift();

    function klass() {
      this.initialize.apply(this, arguments);
    }

    Object.extend(klass, Class.Methods);
    klass.superclass = parent;
    klass.subclasses = [];

    if (parent) {
      var subclass = function() { };
      subclass.prototype = parent.prototype;
      klass.prototype = new subclass;
      parent.subclasses.push(klass);
    }

    for (var i = 0; i < properties.length; i++)
      klass.addMethods(properties[i]);

    if (!klass.prototype.initialize)
      klass.prototype.initialize = Prototype.emptyFunction;

    klass.prototype.constructor = klass;

    return klass;
  },

    isClass: function(object) {
        return (object instanceof Function) && (object.superclass || object === Object);
    }

};

Object.extend(Function.prototype, {

    inspect: function() {
        // Print method name (if any) and the first 80 characters of the decompiled source (without 'function')
        var methodBody = this.toString();
        methodBody = methodBody.substring(8, 88) + (methodBody.length > 88 ? '...' : '');
        return this.qualifiedMethodName() + methodBody;
    },

    qualifiedMethodName: function() {
        return (this.declaredClass ? this.declaredClass + "." : "") + (this.methodName || "anonymous");
    },

    addMethods: function(source) {
        // copy all the methods and properties from {source} into the
        // prototype property of the receiver, which is intended to be
        // a class constructor.  Method arguments named '$super' are treated
        // specially, see Prototype.js documentation for "Class.create()" for details.
        // derived from Class.Methods.addMethods() in prototype.js
        var ancestor = this.superclass && this.superclass.prototype;

        for (var property in source) {
            var value = source[property];
            if (ancestor && Object.isFunction(value) &&
// AT: Qt-specific code -- RegExp support is non-standard in Qt
//                value.argumentNames().first() == "$super") {
                value.firstArgumentName() == "$super") {
                var method = value;
                var advice = (function(m) {
                    return function() { 
                        try { 
                            return ancestor[m].apply(this, arguments) 
                        } catch (e) { 
                            print("Problem with ancestor");
// AT:
//                            console.log("problem with ancestor %s.%s(%s):%s",
//                                Object.inspect(ancestor), m, $A(arguments), e);
//                            Function.showStack();
                            throw e;
                        }
                    };
                })(property);
                advice.methodName = "$superAdvice::" + property;

                value = Object.extend(advice.wrap(method), {
                    valueOf:  function() { return method },
                    toString: function() { return method.toString() },
                    originalFunction: method
                });
            }

            this.prototype[property] = value;
            if (Object.isFunction(value)) {
                for ( ; value; value = value.originalFunction) {
//                    if (value.methodName) {
                        //console.log("class " + this.prototype.constructor.type 
                        // + " borrowed " + value.qualifiedMethodName());
//                    }
                    value.declaredClass = this.prototype.constructor.type;
                    value.methodName = property;
/* AT: Removed Jan 23, 2009
                    if (!this.prototype.constructor.type) {
                        console.log("named " + value.qualifiedMethodName());
                    }
*/
                }
            }
        }

        return this;
    },

    subclass: function(className /*:String*/ /*,... */) {
        // Main method of the LK class system.

        // {className} is the name of the new class constructor which this method synthesizes
        // and binds to {className} in the Global namespace. 
        // Remaining arguments are (inline) properties and methods to be copied into the prototype 
        // of the newly created constructor.

        // modified from prototype.js

        var properties = $A(arguments);
        properties.shift();

        function klass() {
            this.initialize.apply(this, arguments);
        }

        klass.superclass = this;
        klass.subclasses = [];

        var protoclass = function() { }; // that's the constructor of the new prototype object
        protoclass.prototype = this.prototype;
        klass.prototype = new protoclass();

        if (this.subclasses) this.subclasses.push(klass);

        klass.prototype.constructor = klass;
        // .name would be better but js ignores .name on anonymous functions
        klass.type = className;

        for (var i = 0; i < properties.length; i++) {
            klass.addMethods(properties[i] instanceof Function ? (properties[i])() : properties[i]);
        }

        if (!klass.prototype.initialize) {
            klass.prototype.initialize = function() {}; // WAS: Functions.Empty;
        }

        Global[className] = klass;
        return klass;
    },

    getOriginal: function() {
        // get the original 'unwrapped' function, traversing as many wrappers as necessary.
        var func = this;
        while (func.originalFunction) func = func.originalFunction;
        return func;
    }

});

Object.subclasses = [];

// AT: Qt-specific code: showStack is not supported by QtScript
if (!Function.showStack) {
//    print("showStack not supported");
    Function.showStack = function() { /* print("Stack trace unavailable.") */ }; 
}

/*==============================================================================
 * The Prototype Hash class (added from Prototype.js 1.6.0.2 Jan 23, 2009)
 *============================================================================*/

function $H(object) {
  return new Hash(object);
};

var Hash = Class.create(Enumerable, (function() {

  function toQueryPair(key, value) {
    if (Object.isUndefined(value)) return key;
    return key + '=' + encodeURIComponent(String.interpret(value));
  }

  return {
    initialize: function(object) {
      this._object = Object.isHash(object) ? object.toObject() : Object.clone(object);
    },

    _each: function(iterator) {
      for (var key in this._object) {
        var value = this._object[key], pair = [key, value];
        pair.key = key;
        pair.value = value;
        iterator(pair);
      }
    },

    set: function(key, value) {
      return this._object[key] = value;
    },

    get: function(key) {
      return this._object[key];
    },

    unset: function(key) {
      var value = this._object[key];
      delete this._object[key];
      return value;
    },

    toObject: function() {
      return Object.clone(this._object);
    },

    keys: function() {
      return this.pluck('key');
    },

    values: function() {
      return this.pluck('value');
    },

    index: function(value) {
      var match = this.detect(function(pair) {
        return pair.value === value;
      });
      return match && match.key;
    },

    merge: function(object) {
      return this.clone().update(object);
    },

    update: function(object) {
      return new Hash(object).inject(this, function(result, pair) {
        result.set(pair.key, pair.value);
        return result;
      });
    },

    toQueryString: function() {
      return this.map(function(pair) {
        var key = encodeURIComponent(pair.key), values = pair.value;

        if (values && typeof values == 'object') {
          if (Object.isArray(values))
            return values.map(toQueryPair.curry(key)).join('&');
        }
        return toQueryPair(key, values);
      }).join('&');
    },

    inspect: function() {
      return '#<Hash:{' + this.map(function(pair) {
        return pair.map(Object.inspect).join(': ');
      }).join(', ') + '}>';
    },

    toJSON: function() {
      return Object.toJSON(this.toObject());
    },

    clone: function() {
      return new Hash(this);
    }
  }
})());

Hash.prototype.toTemplateReplacements = Hash.prototype.toObject;
Hash.from = $H;

/*======================================================================
 * New version of Lively for the Qt platform ("Lively Qt")
 *====================================================================*/

// The following three variables refer to Qt-specific 
// global structures related to the main window

// These variables are here mainly to facilitate debugging
// They could be removed later

var GlobalWindow; // The main window
var GlobalView;   // The QGraphicsView inside the window
var GlobalScene;  // The QGraphicsScene inside the view

// Pointer to global JavaScript namespace
var Global = this;

/*======================================================================
 * Class QtColor
 * Convenience class that holds preinstantiated QColor objects 
 *====================================================================*/

Object.subclass("QtColor", {
    documentation: "Preinstantiated color objects"
});

// NOTE: QtColor is a singleton class -- do not instantiate!
Object.extend(QtColor, {

    // See "Predefined Colors" in "http://doc.trolltech.com/4.5-snapshot/qcolor.html"
    white:       new QColor(Qt.white),
    black:       new QColor(Qt.black),
    red:         new QColor(Qt.red),
    darkRed:     new QColor(Qt.darkRed),
    green:       new QColor(Qt.green),
    darkGreen:   new QColor(Qt.darkGreen),
    blue:        new QColor(Qt.blue),
    darkBlue:    new QColor(Qt.darkBlue),
    cyan:        new QColor(Qt.cyan),
    darkCyan:    new QColor(Qt.darkCyan),
    magenta:     new QColor(Qt.magenta),
    darkMagenta: new QColor(Qt.darkMagenta),
    yellow:      new QColor(Qt.yellow),
    darkYellow:  new QColor(Qt.darkYellow),
    gray:        new QColor(Qt.gray),
    darkGray:    new QColor(Qt.darkGray),
    lightGray:   new QColor(Qt.lightGray),
    transparent: new QColor(Qt.transparent),

    defaultGradient: (function() {
        var gradient = new QLinearGradient(0, 0, 100, 100);
        gradient.setColorAt(0, new QColor(Qt.gray));
        gradient.setColorAt(1, new QColor(Qt.transparent));
        gradient.setSpread(QGradient.ReflectSpread);
        return gradient;
    })() 
});

// Convenience functions for creating new QPoint and QRect objects
function qpt(x, y) {
    return new QPoint(x, y);
}

function qptf(x, y) {
    return new QPointF(x, y);
}

function qrect(x, y, width, height) {
    return new QRect(x, y, width, height);
}

function qrectf(x, y, width, height) {
    return new QRectF(x, y, width, height);
}

// Holds the last object that we have manipulated or created from 
// the user interface.  This makes it easy to access that object,
// e.g., from the interactive code evaluator.
var that;

/*======================================================================
 * Class QtMorph
 * The main Morph class for Lively on Qt 
 *====================================================================*/

/* This hash table holds a mapping from native peer graphics objects  */
/* (QGraphicsItems) back to morphs.  The hash table is needed because */
/* there is no reliable way to store the backpointer to the morph     */ 
/* in the peer object itself */
PeerToMorphTable = new Hash();

Object.subclass("QtMorph", {
    documentation: "a QtMorph",

    initialize: function() {
        this.setId(QtMorph.newMorphId());
        this.initializeInteractionProperties();
        return this;
    },

    id: function() {
        return this._id;
    },

    setId: function(value) {
        this._id = value;
        return this;
    },

    getType: function() {
        var ctor = this.constructor.getOriginal();
        if (ctor.type) return ctor.type;
        else return "(type unavailable)";
    },

    toString: function() {
        return this.getType() +  "(#" + this._id + "):" + 
               this.getPos() + ":" + this.rect();
    },
    
    inspect: function() {
        return this.toString();
    },

    // Create another, identical copy
    copy: function() {
        // Default implementation returns null;
        // override in subclasses
        return null;
    }

});

/**
  * QtMorph helper functions (NOTE: These are CLASS methods
  * rather than instance methods.)
  */

Object.extend(QtMorph, {

    // Generate a new, unique ID for a QtMorph    
    newMorphId: (function() {
        var morphCounter = 0;
        return function() {
            return ++ morphCounter;
        }
    })(),

    // Generate new Z-value to manage layering of objects
    newZValue: (function() {
        var zCounter = 0;
        return function(front) {
            if (front) return ++zCounter;
            else return 0;
        }
    })()

});

/**
  * QtMorph visual attributes and methods
  */

QtMorph.addMethods({   

    // Default visual attributes
    fillColor: QtColor.lightGray,
    borderColor: QtColor.black,
    borderWidth: 1,
    opacity: 0.8, // So that we can see thru when dropping objects on others!

    // Set the native "peer" object (a QGraphicsItem)
    setPeer: function(nativeItem) {
        this.peer = nativeItem;

        // Backpointer to morph owning the item
        this.peer.setData(0, this);

        // Make sure the object will be on top 
        this.peer.setZValue(QtMorph.newZValue(true));

        // Set other Qt-specific attributes
        /* These flags don't seem to do anything useful, and some
           of them prevent the demos from working properly, e.g.,
           from receiving keyboard events
        this.peer.setFlag(QGraphicsItem.ItemIsMovable,    true);
        this.peer.setFlag(QGraphicsItem.ItemIsSelectable, true);
        this.peer.setFlag(QGraphicsItem.ItemIsFocusable,  true);
        this.peer.setFlag(QGraphicsItem.ItemClipsToShape, true);
        */

        return this;
    },

    // Copy the visual attributes of the object
    // (fill color, border color/width, opacity)
    // to another object
    copyVisualPropertiesTo: function(other) {
        if (!other || !other.peer) return this;
        other.fillColor = this.fillColor;
        other.borderWidth = this.borderWidth;
        other.borderColor = this.borderColor;
        other.opacity = this.getOpacity();
        other.setNativePenAndBrushAttributes();
        return this;
    },

    // This function is called when the visual attributes of 
    // a Lively object change.  It sets the Qt Pen and Brush
    // attributes correctly
    setNativePenAndBrushAttributes: function() {

        if (this.peer) {
            var pen = new QPen();
            var brush;

            // Process the various drawing attributes from Lively
            if (this.fillColor) {
                brush = new QBrush(this.fillColor);
            } else {
                brush = new QBrush(Qt.NoBrush);
            }

             // AT: Qt versions earlier than 4.5 do not support setOpacity
            if (this.opacity && this.peer.setOpacity) {
                this.peer.setOpacity(this.opacity);

                // Warning: This feature is available only in Qt 4.5 and later:
                // Make sure the item does not propagate its transparency to its children
                this.peer.setFlag(QGraphicsItem.ItemDoesntPropagateOpacityToChildren, true);
                this.peer.setFlag(QGraphicsItem.ItemIgnoresParentOpacity, true);
            }

            // Brush (fill) attributes are now complete
            this.peer.setBrush(brush);

            if (this.borderColor) {
                pen.setColor(this.borderColor);
            } else {
                this.peer.setPen(new QPen(Qt.NoPen));
                return; 
            }

            if (this.borderWidth !== undefined) {
                if (this.borderWidth > 0) pen.setWidth(this.borderWidth);
                else {
                    this.peer.setPen(new QPen(Qt.NoPen));
                    return;
                }
            }

            if (this.lineJoin) {
                var QtLineJoinType = null;

                // Map DOM lineJoin types to Qt
                switch (this.lineJoin) {
                case "Miter":
                    QtLineJoinType = Qt.MiterJoin;
                    break;
                case "Round":
                    QtLineJoinType = Qt.RoundJoin;
                    break;
                case "Bevel":
                    QtLineJoinType = Qt.BevelJoin;
                    break;
                default:
                    print("Unrecognized lineJoin type given");
                } 

                if (QtLineJoinType) pen.setJoinStyle(QtLineJoinType);
            }

            if (this.lineCap) {
                var QtLineCapType = null;

                // Map DOM lineJoin types to Qt
                switch (this.lineJoin) {
                case "Butt":
                    QtLineCapType = Qt.FlatCap;
                    break;
                case "Round":
                    QtLineCapType = Qt.RoundCap;
                    break;
                case "Square":
                    QtLineCapType = Qt.SquareCap;
                    break;
                default:
                    print("Unrecognized lineCap type given");
                } 

                if (QtLineCapType) pen.setCapStyle(QtLineCapType);
            }

            // Pen (border) attributes are now complete
            this.peer.setPen(pen);
        }

    },

    setFillColor: function(color) {
        this.fillColor = color;
        this.setNativePenAndBrushAttributes();
        return this;
    },

    getFillColor: function() {
        var brush = this.peer ? this.peer.brush() : 0; 
        // Check first if there's a gradient color associated with the peer
        this.fillColor = brush ? brush.gradient() : 0;
        if (this.fillColor) return this.fillColor;
        // If no gradient, return ordinary color
        this.fillColor = brush ? brush.color() : 0;
        return this.fillColor;
    },

    setBorderColor: function(color) {
        this.borderColor = color;
        this.setNativePenAndBrushAttributes();
        return this;
    },

    getBorderColor: function() {
        var pen = this.peer ? this.peer.pen() : 0; 
        this.borderColor = pen ? pen.color() : 0; 
        return this.borderColor;
    },

    setBorderWidth: function(width) {
        this.borderWidth = width;
        this.setNativePenAndBrushAttributes();
        return this;
    },

    getBorderWidth: function() {
        var pen = this.peer ? this.peer.pen() : 0; 
        this.borderWidth = pen ? pen.width() : 0;
        return this.borderWidth;
    },

    setOpacity: function(alpha) {
        this.opacity = alpha;
        this.setNativePenAndBrushAttributes();
        return this;
    },

    getOpacity: function(alpha) {
        this.opacity = (this.peer && this.peer.opacity) ? this.peer.opacity() : 1;
        return this.opacity;
    },

    // TODO: Line joins and caps are not really fully supported yet
    setLineJoin: function(joinType) {
        if (!joinType) throw new Error("undefined joinType");
        this.lineJoin = joinType;
        this.setNativePenAndBrushAttributes();
        return this;
    },

    setLineCap: function(capType) {
        if (!capType) throw new Error("undefined capType");
        this.lineCap = capType;
        this.setNativePenAndBrushAttributes();
        return this;
    }

});

/**
  * QtMorph interaction properties: these properties indicated
  * whether the object can be moved, copied, resized, rotated, etc.
  * interactively from the user interface.
  * 
  * NOTE: These features only affect the user interface behavior.
  * Objects can still be moved/copied/resized/etc. programmatically
  * regardless of the values of these properties!           
  */

QtMorph.addMethods({

    initializeInteractionProperties: function() {
        // Default values for the interaction properties
        this.allowsMoving = true;
        this.allowsCopying = true;
        this.allowsResizing = true;
        this.allowsRotation = true;
        this.allowsScaling = true;
        this.allowsShearing = true;
        this.allowsAttachingToOthers = true;
        this.allowsAttachingOfOthers = true;
        return this;
    },

    copyInteractionPropertiesTo: function(other) {
        other.allowsMoving   = this.allowsMoving;
        other.allowsCopying  = this.allowsCopying;
        other.allowsResizing = this.allowsResizing;
        other.allowsRotation = this.allowsRotation;
        other.allowsScaling  = this.allowsScaling;
        other.allowsShearing = this.allowsShearing;

        other.allowsAttachingToOthers = this.allowsAttachingToOthers;
        other.allowsAttachingOfOthers = this.allowsAttachingOfOthers;

        return this;
    },
    
    setAllowsMoving: function(state) {
        this.allowsMoving = state;
        return this;
    },

    setAllowsCopying: function(state) {
        this.allowsCopying = state;
        return this;
    },

    setAllowsResizing: function(state) {
        this.allowsResizing = state;
        return this;
    },

    setAllowsRotation: function(state) {
        this.allowsRotation = state;
        return this;
    },

    setAllowsScaling: function(state) {
        this.allowsScaling = state;
        return this;
    },

    setAllowsShearing: function(state) {
        this.allowsShearing = state;
        return this;
    },

    setAllowsAttachingToOthers: function(state) {
        this.allowsAttachingToOthers = state;
        return this;
    },

    setAllowsAttachingOfOthers: function(state) {
        this.allowsAttachingOfOthers = state;
        return this;
    }

});

/**
  * QtMorph transformations: position, bounds, rotation, scaling, etc.
  */

QtMorph.addMethods({

    // Copy this object's location and transformation
    // information to another object
    copyTransformationsTo: function(other) {
        if (!this.peer || !other || !other.peer) return this;
        other.setPos(this.getPos());
        other.peer.setTransform(this.peer.transform());
        return this;
    },

    // Returns a QRectF for the bounds rectangle
    rect: function() {
        return this.peer ? this.peer.rect() : qrect(0, 0, 0, 0);
    },

    // Change the bounds (size) of the object
    // Parameter: QRectF
    setRect: function(rect) {
        if (this.peer) this.peer.setRect(rect);
        return this;
    },

    // Parameter qPoint: a QPoint object
    setPos: function(qPoint) {
        if (this.peer) this.peer.setPos(qPoint);
        return this;
    },

    // Returns a QPointF    
    getPos: function() {
        return this.peer ? this.peer.pos() : qpt(0, 0);
    },

    // NOTE: 'moveTo' and 'moveBy' take parameters as numbers -- not as points!    
    moveTo: function(x, y) {
        this.setPos(qpt(x, y));
        return this;
    },

    moveBy: function(diffX, diffY) {
        if (this.peer) this.peer.moveBy(diffX, diffY);
        return this;
    },

    rotate: function(angle) {
        if (this.peer) this.peer.rotate(angle);
        return this;
    },

    scale: function(scaleX, scaleY) {
        if (this.peer) this.peer.scale(scaleX, scaleY);
        return this;
    },

    shear: function(shearX, shearY) {
        if (this.peer) this.peer.shear(shearX, shearY);
        return this;
    },

    resetTransformations: function() {
        if (this.peer) this.peer.resetTransform();
        return this;
    }

});

/**
  * QtMorph submorph management: adding, removing, enumerating
  */

QtMorph.addMethods({

    // Copy this object's children to other morph.
    // NOTE (possible TODO): This operation does not adjust the JavaScript
    // pointers copied from the the original object to point to the 
    // corresponding copied objects in the other object.
    copyChildrenTo: function(other) {
        if (!other || !other.peer) return this;

        var items = this.peer ? this.peer.childItems() : [];
        for (var i = 0; i < items.length; i++) {
            var item = items[i];
            var morph = PeerToMorphTable.get(item.data(0)); // Not very efficient...
            if (!morph || !morph.copy) continue;
            other.addMorph(morph.copy());
        }
        
        return this;
    },
    
    addMorph: function(morph) { 
        return this.addMorphFrontOrBack(morph, true); 
    },

    addMorphFront: function(morph) {
        return this.addMorphFrontOrBack(morph, true);
      },

    addMorphBack: function(morph) { 
        return this.addMorphFrontOrBack(morph, false);
    },

    addMorphFrontOrBack: function(m, front) {

        // Sanity checks
        if (!m || !m.peer || !this.peer) return m;
        if (m == this) return m; // Cannot add morph to itself!

        // This code is unnecessary; 'setParentItem' does the necessary thing
        // var scene = this.peer.scene();
        // if (!scene) return m;
        // if (!this.peer.scene()) GlobalScene.addItem(m.peer);

        m.peer.setParentItem(this.peer);

        // Add the morph to front (top) or back (bottom) of the display (in Z-order)
        m.peer.setZValue(QtMorph.newZValue(front)); // Larger value = higher in the drawing order

        PeerToMorphTable.set(m.peer.data(0), m);

        return m;
    },

    removeMorph: function(morph) {
        if (morph && morph.peer && this.peer) {
            var scene = this.peer.scene();
            if (!scene) return morph;

            scene.removeItem(morph.peer);
            // TODO: We should remove all the submorphs from the hash table!
            PeerToMorphTable.unset(morph.peer.data(0));
        }
        return morph;
    },

    remove: function() {
        if (!this.peer) return this;

        var scene = this.peer.scene();
        if (!scene) return this; // already removed

        scene.removeItem(this.peer);
        // TODO: We should remove all the submorphs from the hash table!
        PeerToMorphTable.unset(this.peer.data(0));

        return this;
    },

    removeAllMorphs: function() {
        if (!this.peer) return null;

        var scene = this.peer.scene();
        if (!scene) return null;

        var items = this.peer.childItems();
        if (items.length == 0) return null;

        for (var i = 0; i < items.length; i++) {
            var item = items[i];

            scene.removeItem(item);
            PeerToMorphTable.unset(item.data(0));
        }

        // Remove the subworlds also
        var subworlds = this.subworlds;
        if (!subworlds) return;
        
        for (var i = 0; i < subworlds.length; i++) {
            var world = subworlds[i];
            if (world) world.removeAllMorphs();
        }
        subworlds = [];
    },

    // Return all the immediate submorphs of this morph
    submorphs: function() {
        if (!this.peer) return null;

        var items = this.peer.childItems();
        if (items.length == 0) return null;

        var morphs = [];

        for (var i = 0; i < items.length; i++) {
            var item = items[i];

            // Get the corresponding morph
            var morph = PeerToMorphTable.get(item.data(0));
            if (morph) morphs.push(morph);
        }

        return morphs;
    }

});

/**
  * QtMorph event handling support
  */

QtMorph.addMethods({   

    // Initializes the default event handlers for a QtMorph
    setDefaultEventHandlers: function() {
        var item = this.peer;
        if (item) {
            item.mousePressEvent = this.mousePressEvent;
            item.mouseMoveEvent = this.mouseMoveEvent;
            item.mouseReleaseEvent = this.mouseReleaseEvent;
            item.mouseDoubleClickEvent = this.mouseDoubleClickEvent;
            item.contextMenuEvent = this.contextMenuEvent;

            // Currently we use mouse wheel events only for worlds
            // item.wheelEvent = this.wheelEvent;            

            item.keyPressEvent = this.keyPressEvent;
            item.keyReleaseEvent = this.keyReleaseEvent;
        }
        return this;
    },

    // VERY IMPORTANT: The event handling functions below are executed
    // in the context of the peer, i.e., 'this' pointer points 
    // to the peer (QGraphicsItem) rather than to the QtMorph!
    
    // If the UI has been locked, all the event handlers are disabled

    mousePressEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;
        
        // Process only left mouse button clicks here
        // (right button is reserved for contextMenuEvents)
        if ((event.buttons() & Qt.LeftButton) == 0) {
            QtEvents.leftMouseButtonPressed = false;
            return;
        }

        // Bring the item to the top and give it focus
        this.setZValue(QtMorph.newZValue(true));
        this.setFocus();
//        this.scene().setActiveWindow(this);

        // Store a pointer to the morph that was clicked
        QtEvents.currentMorph = PeerToMorphTable.get(this.data(0));
        // print("Mouse pressed on: " + QtEvents.currentMorph + ", event: " + event);

        // Make the morph available to the programmer via 'that'
        if (QtEvents.currentMorph)
             that = QtEvents.currentMorph;
        else that = null;

        // Store the location of the event
        QtEvents.lastMouseDownPos = qpt(event.scenePos().x(), event.scenePos().y());

        // Record the difference between the mouse click 
        // and the position of the object; we need the diff
        // info, e.g., when moving the object
        QtEvents.eventPosDiff = qpt(event.scenePos().x(), event.scenePos().y());
        QtEvents.eventPosDiff.operator_subtract_assign(this.scenePos());

        QtEvents.leftMouseButtonPressed = true;
        QtEvents.mouseWasDragged = false;

        // Check if Shift key is down, i.e., the user wants to copy the object
        if (event.modifiers() & Qt.ShiftModifier) {
            // Copy the current morph
            var morph = QtEvents.currentMorph;
            if (morph) {
                var newMorph = morph.copy();
                var currentWorld = QtWorldMorph.current();
                if (newMorph && currentWorld) {
                    currentWorld.addMorph(newMorph);
                    newMorph.setPos(QtEvents.lastMouseDownPos);
                    that = newMorph;
                    QtEvents.currentMorph = null;
                    return; // Done with copying!
                }            
            }
        }

        QtEvents.transformationInProgress = false;
        if (!that) return;

        // Check for transformation requests (resizing, rotation, scaling, ...)
        if (that.allowsResizing && that.mouseOnResizeHandle(event)) {
            QtEvents.transformationInProgress = QtEvents.RESIZING;
            that.setHelpText(event, "Resizing...", 10);
            return;
        }

        if (that.allowsRotation && that.mouseOnRotationHandle(event)) {
            QtEvents.transformationInProgress = QtEvents.ROTATING;
            that.setHelpText(event, "Rotating...", 10);
            return;
        }

        if (that.allowsScaling && that.mouseOnScalingHandle(event)) {
            QtEvents.transformationInProgress = QtEvents.SCALING;
            that.setHelpText(event, "Scaling...", -50);
            return;
        }

        if (that.allowsShearing && that.mouseOnShearingHandle(event)) {
            QtEvents.transformationInProgress = QtEvents.SHEARING;
            that.setHelpText(event, "Shearing...", -60);
            return;
        }

    },

    mouseMoveEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;

        // Process only left mouse button clicks here
        // (right button is reserved for contextMenuEvents)
        if ((event.buttons() & Qt.LeftButton) == 0) return;
        if (!QtEvents.currentMorph) return;

        // print("Mouse moved on: " + QtEvents.currentMorph + ", event: " + event);

        // Check if we are resizing, rotating or scaling the object
        var xPos; // For help text positioning
        switch (QtEvents.transformationInProgress) {

        case QtEvents.RESIZING: 
        {
            var bounds = this.rect();
            var morph = QtEvents.currentMorph;
            if (morph) {
                var newWidth  = event.pos().x() - bounds.x();
                var newHeight = event.pos().y() - bounds.y();

                // Treat ellipses specially
                if (that instanceof QtEllipseMorph) newHeight *= 2;

                // Make sure size does not go negative
                if (newWidth  <= QtEvents.precision) newWidth  = QtEvents.precision;
                if (newHeight <= QtEvents.precision) newHeight = QtEvents.precision;

                var newBounds = qrect(bounds.x(), bounds.y(), newWidth, newHeight);
                morph.setRect(newBounds);
            }
            xPos = 10;
            break;
        }

        case QtEvents.ROTATING:
        {
            var morph = QtEvents.currentMorph;
            if (morph) {
                var mousePos = event.scenePos();
                var diff = mousePos.y() - QtEvents.lastMouseDownPos.y();
                if (diff) { 
                    // TODO: This is just a poor approximation of proper mouse action.
                    // Use trigonometric functions later.
                    morph.rotate(diff);
                    QtEvents.lastMouseDownPos = mousePos;
                }
            }
            xPos = 10;
            break;
        }

        case QtEvents.SCALING: {
            var morph = QtEvents.currentMorph;
            if (morph) {
                var mousePos = event.scenePos();
                var diff = mousePos.y() - QtEvents.lastMouseDownPos.y();
                if (diff) { 
                    // TODO: This is just a poor approximation of proper mouse action.
                    // Use trigonometric functions later.
                    diff /= 20;

                    // Avoid growing or shrinking the object too quickly
                    diff += 1;
                    if (diff > 2) diff = 2;
                    if (diff < 0.5) diff = 0.5;

                    morph.scale(diff, diff);
                    QtEvents.lastMouseDownPos = mousePos;
                }
            }
            xPos = -50;
            break;
        }

        case QtEvents.SHEARING: {
            var morph = QtEvents.currentMorph;
            if (morph) {
                var mousePos = event.scenePos();
                var diffX = mousePos.x() - QtEvents.lastMouseDownPos.x();
                var diffY = mousePos.y() - QtEvents.lastMouseDownPos.y();
                if (diffX || diffY) { 
                    // TODO: This is just a poor approximation of proper mouse action.
                    // Use trigonometric functions later.
                    diffX /= 100;
                    diffY /= 100;

                    morph.shear(diffX, diffY);
                    QtEvents.lastMouseDownPos = mousePos;
                }
            }
            xPos = -60;
            break;
        }}

        // If help text has been set, move the text with the cursor
        var help = QtEvents.helpTextItem; 
        if (help) {
            var helpPos = qpt(event.scenePos().x() + xPos, event.scenePos().y() - 10);
            help.setPos(helpPos);            
        }

        if (QtEvents.transformationInProgress) return;

        // If we are here, we are moving the object, rather than 
        // performing a transformation.

        // Make sure the objects allows itself to be moved
        if (!QtEvents.currentMorph.allowsMoving) return;
        
        // Check for "wormholes": special QtProxyMorphs that allow objects
        // to be transported between worlds
        if (QtEvents.wormholesEnabled) {

            var scene = this.scene();
            var itemsBelow = scene ? scene.items(event.scenePos()) : null;
            var secondTopmostItem = itemsBelow ? itemsBelow[1] : null;
            if (secondTopmostItem) {
                var secondTopmostMorph = PeerToMorphTable.get(secondTopmostItem.data(0));
                if (secondTopmostMorph && secondTopmostMorph.wormhole) {
                    var otherWorld = secondTopmostMorph.wormhole;
                    if (otherWorld && (otherWorld.peer != this.parentItem())) {
                        QtEvents.currentMorph.remove();
                        otherWorld.addMorph(QtEvents.currentMorph);
                        otherWorld.setCurrent();
                    }

                    // Wormhole transport completed!
                    return;
                }
            } 
        }

        // Perform an ordinary move

        // Cleared for moving...
        QtEvents.mouseWasDragged = true;

        // While moving, the object must be reparented to the world.
        // Otherwise, 'setPos' will not do the expected thing.
        var currentWorld = QtWorldMorph.current();
        if (this.parentItem() != currentWorld.peer) this.setParentItem(currentWorld.peer);

        // Move the object & adjust position based on 'eventPosDiff'
        var eventPos = qpt(event.scenePos().x(), event.scenePos().y());
        eventPos.operator_subtract_assign(QtEvents.eventPosDiff);
        this.setPos(eventPos);
    },

    mouseReleaseEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;

        // NOTE: The test below does not work because the mouse is already up
        // at this point; we must use a variable to filter out right clicks
        // if ((event.buttons() & Qt.LeftButton) == 0) return;
        if (!QtEvents.leftMouseButtonPressed) return;

        // print("Mouse released on: " + QtEvents.currentMorph + ", event: " + event);
        QtEvents.currentMorph = null;

        // Get pointer to the QGraphicsScene (must exist or we wouldn't even be here!)
        var scene = this.scene();

        // If help text has been set, remove the "floating" text now
        var help = QtEvents.helpTextItem; 
        if (help) {
            scene.removeItem(help);
            help.setParentItem(null);
            QtEvents.helpTextItem = null;
        }

        // Do not reparent the object if the mouse did not move at all.
        // TODO: We should use 'lastMouseDownPos' to allow some mouse movement
        // (e.g., for pen-based devices in which pen coordinates are not precise)
        if (!QtEvents.mouseWasDragged) return;

        // If this was a resize/rotation/scaling operation, do not reparent
        if (QtEvents.transformationInProgress) {
            QtEvents.transformationInProgress = false;
            return;
        }

        var itemsBelow = scene.items(event.scenePos());

        // NOTE: 'itemsBelow' contains the items under mouse in Z order
        // Index 0 is the item that we are moving; index 1 is the new parent

        // If the topmost object under mouse is not a QtMorph (e.g., the mouse
        // is on a QWidget, we ignore the reparenting request (the morph remains 
        // attached to the world; see mouseMoveEvent)
        var topmostItem = itemsBelow ? itemsBelow[0] : null;
        if (!topmostItem) return;

        var topmostMorph = PeerToMorphTable.get(topmostItem.data(0)); 
        if (!topmostMorph || topmostMorph !== that || !that.allowsAttachingToOthers) return;

        // If we got here, we are carrying 'that' morph

        // New parent item is the one right below the topmost item
        var newParentItem = itemsBelow[1];

        // Nothing to drop on (sanity check; should not occur
        // since the world item should always be available below)
        if (!newParentItem) return;

        // If the receiver is not a QtMorph, we ignore the reparenting request
        // (the morph remain attached to the world because of mouseMoveEvent)
        var newParentMorph = PeerToMorphTable.get(newParentItem.data(0)); 
        if (!newParentMorph || !newParentMorph.allowsAttachingOfOthers) return;

        // Change ownership and adjust object position accordingly
        if (newParentItem != this.parentItem()) {
            this.setParentItem(newParentItem);

            var newPos = qpt(this.pos().x(), this.pos().y());
            newPos.operator_subtract_assign(newParentItem.scenePos());
            this.setPos(newPos);
        }
    },

    mouseDoubleClickEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;

        this.setFocus();

        // Store a pointer to the morph that was clicked
        QtEvents.currentMorph = PeerToMorphTable.get(this.data(0));

        // Default action: open a popup menu
        this.contextMenuEvent(event);

        // print("Mouse double clicked on: " + QtEvents.currentMorph + ", event: " + event);
    },

    wheelEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;

        // We currently use wheel events to zoom the world
        if (QtWorldMorph.wheelZoomingEnabled) {
            var currentWorld = QtWorldMorph.current();

            var wheelDelta = event.delta() / 120;

            if (wheelDelta < 0) {
                if (currentWorld.peer) currentWorld.peer.scale(0.8, 0.8);
            } else {
                if (currentWorld.peer) currentWorld.peer.scale(1.25, 1.25);
            }
        }

        // print("Wheel event on: " + QtEvents.currentMorph + ", event: " + event);
    },

    contextMenuEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;

        // Store a pointer to the morph that was clicked
        var currentMorph = QtEvents.currentMorph = PeerToMorphTable.get(this.data(0));

        // print("Context menu event on: " + QtEvents.currentMorph + ", event: " + event);

        // Make the event position accessible to the function(s) called below
        QtEvents.lastMouseDownPos = qpt(event.scenePos().x(), event.scenePos().y());
        QtEvents.leftMouseButtonPressed = false;
        QtEvents.mouseWasDragged = false;
        QtEvents.transformationInProgress = false;

        var menu = new QMenu();
            var action = menu.addAction("Copy");
            action.triggered.connect(this, QtEvents.copyMorphToWorld);
            if (!currentMorph || !currentMorph.allowsCopying) action.enabled = false;
            menu.addAction("Remove").triggered.connect(this, QtEvents.removeMorphFromWorld);
        var sub1 = menu.addMenu("Style")
            sub1.addAction("Fill Color...").triggered.connect(this, QtEvents.changeFillColor);
            sub1.addAction("Border Color...").triggered.connect(this, QtEvents.changeBorderColor);
            var sub11 = sub1.addMenu("Border Width");
                sub11.addAction("0").triggered.connect(this, QtEvents.changeBorderWidthTo0);
                sub11.addAction("1").triggered.connect(this, QtEvents.changeBorderWidthTo1);
                sub11.addAction("2").triggered.connect(this, QtEvents.changeBorderWidthTo2);
                sub11.addAction("3").triggered.connect(this, QtEvents.changeBorderWidthTo3);
                sub11.addAction("4").triggered.connect(this, QtEvents.changeBorderWidthTo4);
                sub11.addAction("5").triggered.connect(this, QtEvents.changeBorderWidthTo5);
                sub11.addAction("6").triggered.connect(this, QtEvents.changeBorderWidthTo6);
            var sub12 = sub1.addMenu("Opacity");
                sub12.addAction("5%").triggered.connect(this, QtEvents.changeOpacityTo5);
                sub12.addAction("20%").triggered.connect(this, QtEvents.changeOpacityTo20);
                sub12.addAction("40%").triggered.connect(this, QtEvents.changeOpacityTo40);
                sub12.addAction("60%").triggered.connect(this, QtEvents.changeOpacityTo60);
                sub12.addAction("80%").triggered.connect(this, QtEvents.changeOpacityTo80);
                sub12.addAction("100%").triggered.connect(this, QtEvents.changeOpacityTo100);
        if (QtEvents.currentMorph instanceof QtProxyMorph) {
            var sub13 = sub1.addMenu("Widget Inset");
                sub13.addAction("0").triggered.connect(this, QtEvents.changeInsetTo0);
                sub13.addAction("2").triggered.connect(this, QtEvents.changeInsetTo2);
                sub13.addAction("4").triggered.connect(this, QtEvents.changeInsetTo4);
                sub13.addAction("6").triggered.connect(this, QtEvents.changeInsetTo6);
                sub13.addAction("8").triggered.connect(this, QtEvents.changeInsetTo8);
                sub13.addAction("10").triggered.connect(this, QtEvents.changeInsetTo10);
        }
        var sub2 = menu.addMenu("Transform");
            sub2.addAction("Shrink").triggered.connect(this, QtEvents.shrinkByTwo);
            sub2.addAction("Enlarge").triggered.connect(this, QtEvents.enlargeByTwo);
            sub2.addAction("Rotate Left").triggered.connect(this, QtEvents.rotateLeft);
            sub2.addAction("Rotate Right").triggered.connect(this, QtEvents.rotateRight);
            sub2.addAction("Shear Left").triggered.connect(this, QtEvents.shearLeft);
            sub2.addAction("Shear Right").triggered.connect(this, QtEvents.shearRight);
            sub2.addSeparator();
            sub2.addAction("Reset Transformations").triggered.connect(this, QtEvents.resetTransformations);
        menu.addAction("Properties...").triggered.connect(this, QtEvents.showPropertiesDialog);
        menu.addSeparator();
        menu.addAction("Inspect").triggered.connect(this, QtEvents.inspectThisObject);
        menu.addAction("Show Tooltip").triggered.connect(this, QtEvents.whoAmI);
        menu.popup(QCursor.pos());
    },

    keyPressEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;        
        // print("Key pressed on: " + this + ", event: " + event);
    },

    keyReleaseEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;
        // print("Key released on: " + this + ", event: " + event);
    }

});

/**
  * QtMorph handle checking for transformations (resizing, rotation, scaling, ...)
  */

QtMorph.addMethods({

    mouseOnResizeHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((bounds.x() + bounds.width()  - pos.x() <= precision) && 
            (bounds.y() + bounds.height() - pos.y() <= precision)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(bounds.x() + bounds.width()  - precision, 
               //              bounds.y() + bounds.height() - precision, 
               //              precision, precision);
               return true;
        } else return false;
    },

    mouseOnRotationHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((bounds.x() + bounds.width() - pos.x() <= precision) && 
            (pos.y() - bounds.y()                  <= precision)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(bounds.x() + bounds.width() - precision, bounds.y(), 
               //              precision, precision);
               return true;
        } else return false;
    },

    mouseOnScalingHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((pos.x() - bounds.x()                   <= precision) && 
            (bounds.y() + bounds.height() - pos.y() <= precision)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(bounds.x(), bounds.y() + bounds.height() - precision, 
               //              precision, precision);
               return true;
        } else return false;
    },

    mouseOnShearingHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((pos.x() - bounds.x() <= precision) &&
            (pos.y() - bounds.y() <= precision)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(0, 0, precision, precision);
               return true;
        } else return false;
    },

    // Display a "floating" help text while performing an interactive transformation
    // TODO: When clicking near the left edge of an object,
    // help text should be placed on the left side of the object!
    setHelpText: function(event, string, xPos) {
        var help = QtEvents.helpTextItem = new QGraphicsSimpleTextItem(string); 
        help.setParentItem(QtWorldMorph.current().peer);
        var helpPos = qpt(event.scenePos().x() + xPos, event.scenePos().y() - 10);
        help.setPos(helpPos);
        help.setZValue(QtMorph.newZValue(true));    
    }

});

/*======================================================================
 * Class QtRectangleMorph
 * Simple rectangle morph 
 *====================================================================*/

QtMorph.subclass("QtRectangleMorph", {

    documentation: "a QtRectangleMorph",

    // Parameter rect: a QRect object (defines the bounds of the rectangle)
    // Parameter fillColor: a QColor object
    // Parameter borderWidth: a Number
    // Parameter borderColor: a QColor object
    // If any of the parameters is null/undefined, we use default values
    initialize: function($super, rect, fillColor, borderWidth, borderColor) {
        $super();

        // TODO: Default shape should be fully rectangular rather than oblong?
        var item = rect ? new QGraphicsRectItem(rect.x(), rect.y(), rect.width(), rect.height(), null) 
                        : new QGraphicsRectItem(0, 0, 80, 40, null);
        this.setPeer(item);

        // NOTE: event handlers must be initialized _after_ setting the peer.
        // Otherwise, event handlers will not be initialized correctly. 
        this.setDefaultEventHandlers();

        if (fillColor) this.fillColor = fillColor;
        if (borderWidth) this.borderColor = borderColor;
        if (borderColor) this.borderWidth = borderWidth; 
        this.setNativePenAndBrushAttributes();

        return this;
    },
    
    copy: function() {
        if (!this.peer) return null;
        
        var other = new QtRectangleMorph(this.rect(),           this.getFillColor(), 
                                         this.getBorderWidth(), this.getBorderColor());

        // Opacity must be copied explicitly
        other.setOpacity(this.getOpacity());

        // Copy location and transformation info
        this.copyTransformationsTo(other);

        // Copy interaction properties
        this.copyInteractionPropertiesTo(other);

        // Copy children as well
        this.copyChildrenTo(other);

        return other;
    }
    
});

/*======================================================================
 * Class QtEllipseMorph
 * Simple ellipse morph 
 *====================================================================*/

QtMorph.subclass("QtEllipseMorph", {

    documentation: "a QtEllipseMorph",

    // Parameter rect: a QRect object (defines the bounds of the ellipse)
    // Parameter fillColor: a QColor object
    // Parameter borderWidth: a Number
    // Parameter borderColor: a QColor object
    // If any of the parameters is null/undefined, we use default values
    initialize: function($super, rect, fillColor, borderWidth, borderColor) {
        $super();

        // TODO: Default shape should be a perfect circle rather than an ellipse?
        var item = rect ? new QGraphicsEllipseItem(rect.x(), rect.y(), rect.width(), rect.height(), null) 
                        : new QGraphicsEllipseItem(-40, -20, 80, 40, null); // Position around ellipse midpoint!
        this.setPeer(item);

        // NOTE: event handlers must be initialized _after_ setting the peer.
        // Otherwise, event handlers will not be initialized correctly. 
        this.setDefaultEventHandlers();

        if (fillColor) this.fillColor = fillColor;
        if (borderWidth) this.borderColor = borderColor;
        if (borderColor) this.borderWidth = borderWidth; 
        this.setNativePenAndBrushAttributes();

        return this;
    },

    copy: function() {
        if (!this.peer) return null;
        
        var other = new QtEllipseMorph(this.rect(),           this.getFillColor(), 
                                       this.getBorderWidth(), this.getBorderColor());

        // Opacity must be copied explicitly
        other.setOpacity(this.getOpacity());

        // Copy location and transformation info
        this.copyTransformationsTo(other);

        // Copy interaction properties
        this.copyInteractionPropertiesTo(other);

        // Copy children as well
        this.copyChildrenTo(other);

        return other;
    },

    // Transformation handles (for resizing, rotation, scaling, ...)
    // are in a different place in ellipse morphs.  We implement the 
    // ellipse-specific checking functions here.

    mouseOnResizeHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((bounds.x() + bounds.width() - pos.x() <= precision) && 
            (Math.abs(bounds.y() + bounds.height()/2 - pos.y()) <= precision/2)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(bounds.x() + bounds.width() - precision, 
               //              bounds.y() + bounds.height()/2 - precision/2, 
               //              precision, precision);
               return true;
        } else return false;
    },
    
    mouseOnRotationHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((pos.y() - bounds.y() <= precision) && 
            (Math.abs(bounds.x() + bounds.width()/2 - pos.x()) <= precision/2)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(bounds.x() + bounds.width()/2 - precision/2, 
               //              bounds.y(), precision, precision);
               return true;
        } else return false;
    },
    
    mouseOnScalingHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((bounds.y() + bounds.height() - pos.y() <= precision) && 
            (Math.abs(bounds.x() + bounds.width()/2 - pos.x()) <= precision/2)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(bounds.x() + bounds.width()/2 - precision/2, 
               //              bounds.y() + bounds.height() - precision, 
               //              precision, precision);
               return true;
        } else return false;
    },
    
    mouseOnShearingHandle: function(event) {
        var bounds = this.rect();
        var pos = event.pos();
        var precision = QtEvents.precision;
        if ((pos.x() - bounds.x() <= precision) &&
            (Math.abs(bounds.y() + bounds.height()/2 - pos.y()) <= precision/2)) {

               // OLD CODE: Return a rectangle representing the handle
               // return qrect(bounds.x(), bounds.y() + bounds.height()/2 - precision/2, 
               //              precision, precision);
               return true;
        } else return false;
    }

});

/*======================================================================
 * Class QtProxyMorph
 * A QtMorph that can host any QWidget
 *====================================================================*/

QtRectangleMorph.subclass("QtProxyMorph", {

    documentation: "a QtProxyMorph",
    
    // The default amount of pixels surrounding the QWidget
    inset: 5,
    fillColor: QtColor.defaultGradient, // Use gradient color by default 

    // Parameter qWidget: a QWidget object instance
    // Parameter rect: a QRect object (defines the bounds of the morph)
    // Parameter inset: The number of pixels allocated around the qWidget
    // Parameters rect and inset are optional; default values are used automatically
    initialize: function($super, qWidget, rect, inset) {

        // IMPORTANT: Must allow 0 as an inset parameter
        if (inset !== undefined) this.inset = inset;
        if (inset < 0) this.inset = 0;

        if (rect) $super(rect, this.fillColor, this.borderWidth, this.borderColor);
        else $super(qrect(0, 0, qWidget.width+this.inset*2, qWidget.height+this.inset*2),
                    this.fillColor, this.borderWidth, this.borderColor);

        // If bounds have been specified, resize the widget accordingly
        if (rect) {
            var bounds = this.rect();
            qWidget.resize(bounds.width()  - this.inset*2, 
                           bounds.height() - this.inset*2);
        }

        // Create a new QGraphicsProxyWidget to host the QWidget
        this.proxyWidget = new QGraphicsProxyWidget();
        this.proxyWidget.setWidget(qWidget);
        this.inner = qWidget;
        this.proxyAsItem = this.castToQGraphicsItem(this.proxyWidget);
        
        this.proxyAsItem.setParentItem(this.peer);
        this.proxyAsItem.setPos(this.rect().x() + this.inset, this.rect().y() + this.inset);
        
        return this;
    },

    // Because of missing multiple inheritance support in JavaScript,
    // we must patch ("cast") the peer pointer by obtaining the same 
    // pointer as a QGraphicsItem from the scene.
    // TODO: Perhaps this operation should be located in QtEvents?
    castToQGraphicsItem: function(proxyWidget) {
        // Attach the item temporarily to the scene
        GlobalScene.addItem(proxyWidget);
        var allItems = GlobalScene.items();
        var widgetAsItem = allItems[allItems.length - 1];
        GlobalScene.removeItem(proxyWidget);

        // Return the real peer pointer
        return widgetAsItem;
    },

    copy: function() {
        if (!this.peer || !this.inner) return null;

        // Create a new QWidget. Unfortunately, we cannot copy the 
        // attributes/parameters of the widget automatically.
        var newQWidget = new this.inner.constructor();
        if (!newQWidget) return null;

        var bounds = this.rect();
        newQWidget.resize(bounds.width()  - this.inset*2, 
                          bounds.height() - this.inset*2);

        var other = new QtProxyMorph(newQWidget, bounds, this.inset);

        // Copy location, transformation, visuals (colors, etc.) and other properties 
        this.copyTransformationsTo(other);
        this.copyVisualPropertiesTo(other);
        this.copyInteractionPropertiesTo(other);

        // Copy children as well
        this.copyChildrenTo(other);
        
        return other;
    },
    
    // Change the bounds (size) of the object
    // Parameter: QRectF
    setRect: function($super, rect) {
        $super(rect);
        if (this.proxyAsItem) this.proxyAsItem.setPos(rect.x() + this.inset, rect.y() + this.inset);
        if (this.inner) this.inner.resize(rect.width()  - this.inset*2,
                                          rect.height() - this.inset*2);
        return this;
    },
    
    // Change the border size (inset) of the object
    setInset: function(newInset) {
        this.inset = (newInset > 0) ? newInset : 0;
        this.setRect(this.rect());
        return this;
    },

    toString: function($super) {
        return $super() + ":" + this.inner;
    }

});

/*======================================================================
 * Class QtWorldMorph
 * A QtMorph representing a world that can host other morphs
 *====================================================================*/

QtMorph.subclass("QtWorldMorph", {
    documentation: "a QtWorldMorph"
});

/**
  * QtWorldMorph class methods and variables
  */

Object.extend(QtWorldMorph, {    
    
    // Default world size
    defaultExtent: { x: 1250, y: 750 },

    // TODO: Implement full timer/scheduler functionality later    
    timer: null,

    // Enables/disables world zooming with wheelEvents
    wheelZoomingEnabled: false,

    // Pointer to the main world (the world that was created first)
    mainWorld: null,

    // Pointer to the currently active, visible world
    currentWorld: null,

    current: function() {
        return QtWorldMorph.currentWorld;
    },

    setCurrent: function(newWorld) {
        if (newWorld && newWorld.peer && newWorld instanceof QtWorldMorph) {
            QtWorldMorph.currentWorld = newWorld;

            // Add the world to a scene (if not already done)
            newWorld.displayWorldOnItsScene();

            // Make sure the world is on top of the Z order 
            newWorld.peer.setZValue(QtMorph.newZValue(true));
        }
    }
    
});

/**
  * QtWorldMorph instance methods
  */

QtWorldMorph.addMethods({

    // Worlds should not be transparent; override the value defined by QtMorph
    opacity: 1,

    // Parameter rect: a QRect object (defines the size of the world)
    // Parameter fillColor: a QColor object
    // Parameter borderWidth: a Number
    // Parameter borderColor: a QColor object
    // If any of the parameters is null/undefined, we use default values
    initialize: function($super, rect, fillColor, borderWidth, borderColor) {
        $super();

        // Create a new graphics scene and view for the world    
        this.createNewSceneAndView(rect);

        // Create the background of the world (a QGraphicsRectItem)
        var item = rect ? new QGraphicsRectItem(rect.x(), rect.y(), rect.width(), rect.height(), null) 
                        : new QGraphicsRectItem(0, 0, QtWorldMorph.defaultExtent.x, QtWorldMorph.defaultExtent.y, null);
        this.setPeer(item);

        // NOTE: The world has its own special event handling support.
        // Do not call QtMorph.setDefaultEventHandlers!
        this.setWorldEventHandlers();

        if (fillColor) this.fillColor = fillColor;
        if (borderWidth) this.borderColor = borderColor;
        if (borderColor) this.borderWidth = borderWidth; 
        this.setNativePenAndBrushAttributes(); // NOTE: Will also set opacity

        // Keep track of the firstborn (main) world
        if (QtWorldMorph.mainWorld == null) QtWorldMorph.mainWorld = this;

        // Each world keeps track of its subworlds
        this.subworlds = [];

        return this;
    },

    // Set the native "peer" object (in this case: a QGraphicsRectItem)
    // NOTE: The world's attributes are different from the default attributes;
    // therefore, we override 'setPeer' here
    setPeer: function(nativeItem) {
        this.peer = nativeItem;

        // Backpointer to morph owning the item
        this.peer.setData(0, this);

        // Ensure the world is always displayed _behind_ everything else 
        this.peer.setZValue(-1);

        // Ensure children will not go beyond the boundaries of the world
        // this.peer.setFlag(QGraphicsItem.ItemClipsToShape, true);

        // NOTE: Qt 4.5 crashes when the following flag is set and proxy widgets are moved!
        // this.peer.setFlag(QGraphicsItem.ItemClipsChildrenToShape, true);
    
        return this;
    },

    // Create the new QGraphicsScene and QGraphicsView associated with this world
    // Parameter rect: a QRect object (defines the size of the world)
    // If rect is null/undefined, we use the default world size
    createNewSceneAndView: function(rect) {
        var scene = new QGraphicsScene(GlobalWindow);
        var view;

        // TODO: Get rid of GlobalScene and GlobalView later
        if (!GlobalScene) GlobalScene = scene;

        // NOTE: Currently we only instantiate one QGraphicsView
        if (GlobalView) {
            view = GlobalView;
        } else {
            view = GlobalView = new QGraphicsView(scene, GlobalWindow);
            view.setRenderHint(QPainter.Antialiasing, true);
            view.setRenderHint(QPainter.TextAntialiasing, true);
            view.setRenderHint(QPainter.SmoothPixmapTransform, true);
        }

        if (rect) {
            view.resize(rect.width(), rect.height());
        } else {
            view.resize(QtWorldMorph.defaultExtent.x, QtWorldMorph.defaultExtent.y);
        }

        // NOTE: In order to avoid scrollbars from being created for the scene,
        // the size of the scene rect must be smaller than the window/view rect
        if (rect) {
            scene.setSceneRect(rect.x()+2, rect.y()+2, rect.width()-4, rect.height()-4);
        } else {
            scene.setSceneRect(2, 2, QtWorldMorph.defaultExtent.x-4, QtWorldMorph.defaultExtent.y-4);
        }
        
        this.scene = scene;
        this.view  = view;

        return this;
    },
    
    // Make the world visible in a QGraphicsScene
    displayWorldOnItsScene: function() {
        if (this.peer) {
            // Make sure the world's peer (the rect item) is added to the scene
            // (and make sure that it is added only once)
            if (!this.peer.scene()) this.scene.addItem(this.peer);

            // If the world is not in PeerToMorphTable already,
            // add it to the table now
            if (!PeerToMorphTable.get(this.peer.data(0))) PeerToMorphTable.set(this.peer.data(0), this);
            
            // This operation performs the actual world (view) switch
            this.view.setScene(this.scene);
            this.scene.setFocus();
            this.view.show();
        }
        return this;
    },

    // Set this world current
    setCurrent: function() {
        QtWorldMorph.setCurrent(this);
        return this;
    }

});
    
/**
  * QtWorldMorph event handling operations
  * 
  * NOTE: Even though QtWorldMorph is a subclass of QtMorph,
  * the world implements only a subset of the event handling
  * operations.  The necessary changes are made here.       
  */

QtWorldMorph.addMethods({

    // Initializes the default event handlers for a QtWorldMorph.
    // NOTE: The world implements only a small subset of QtMorph event handlers!
    setWorldEventHandlers: function() {
        var item = this.peer;
        if (item) {
            item.contextMenuEvent = this.contextMenuEvent;
            item.wheelEvent = this.wheelEvent;

            item.mousePressEvent = this.mousePressEvent;
            item.mouseDoubleClickEvent = this.mouseDoubleClickEvent;

        }
        return this;
    },

    // VERY IMPORTANT: The event handling functions below are executed
    // in the context of the peer, i.e., 'this' pointer points 
    // to the peer (QGraphicsItem) rather than to the QtWorldMorph!

    mousePressEvent: function(event) {
        
        // Allows the user to get a 'that' pointer to the world!
        var currentMorph = PeerToMorphTable.get(this.data(0));

        // To avoid embarrassing moments especially during demos,
        // we never set 'that' pointer to the main world!
        if (currentMorph != QtWorldMorph.mainWorld) that = currentMorph;

        // NOTE: mousePressEvent must be defined, or otherwise
        // the mouseDoubleClickEvent (see 'setWorldEventHandlers'
        // above) would not be delivered to its handler (which is
        // inherited from QtMorph)
    },

    contextMenuEvent: function(event) {
        if (QtEvents.userInterfaceLocked) return;

        // Store a pointer to the morph that was clicked
        QtEvents.currentMorph = PeerToMorphTable.get(this.data(0));

        // print("Context menu event on: " + QtEvents.currentMorph + ", event: " + event);

        // Make the event accessible to the function(s) called below
        QtEvents.lastMouseDownPos = qpt(event.scenePos().x(), event.scenePos().y());
        QtEvents.leftMouseButtonPressed = false;
        QtEvents.mouseWasDragged = false;
        QtEvents.transformationInProgress = false;

        // Open a Qt popup menu
        var menu = new QMenu();
        var sub1 = menu.addMenu("Simple Objects");
            sub1.addAction("Rectangle").triggered.connect(this, QtEvents.newRectangle);
            sub1.addAction("Ellipse").triggered.connect(this, QtEvents.newEllipse);
            sub1.addAction("Polygon").triggered.connect(this, QtEvents.unimplemented);
            sub1.addAction("Polyline").triggered.connect(this, QtEvents.unimplemented);
            sub1.addAction("Simple Text").triggered.connect(this, QtEvents.unimplemented);
        var sub2 = menu.addMenu("Tools");
            sub2.addAction("Class Browser").triggered.connect(this, QtEvents.newClassBrowser);
            sub2.addAction("Inspector").triggered.connect(this, QtEvents.newInspector);
            sub2.addAction("Evaluator").triggered.connect(this, QtEvents.newEditor);
            sub2.addAction("Debugger").triggered.connect(this, QtEvents.openDebugger);
        var sub3 = menu.addMenu("QWidgets");
            sub3.addAction("QCalendarWidget").triggered.connect(this, QtEvents.newQCalendarWidget);
            sub3.addAction("QCheckBox").triggered.connect(this, QtEvents.newQCheckBox);
            sub3.addAction("QColorDialog").triggered.connect(this, QtEvents.newQColorDialog);
            sub3.addAction("QComboBox").triggered.connect(this, QtEvents.newQComboBox);
            sub3.addAction("QDateTimeEdit").triggered.connect(this, QtEvents.newQDateTimeEdit);
            sub3.addAction("QDial").triggered.connect(this, QtEvents.newQDial);
            sub3.addAction("QLabel (QImage)").triggered.connect(this, QtEvents.newQLabel);
            sub3.addAction("QLineEdit").triggered.connect(this, QtEvents.newQLineEdit);
            sub3.addAction("QProgressBar").triggered.connect(this, QtEvents.newQProgressBar);
            sub3.addAction("QPushButton").triggered.connect(this, QtEvents.newQPushButton);
            sub3.addAction("QScrollBar").triggered.connect(this, QtEvents.newQScrollBar);
            sub3.addAction("QSlider").triggered.connect(this, QtEvents.newQSlider);
            sub3.addAction("QSpinBox").triggered.connect(this, QtEvents.newQSpinBox);
            sub3.addAction("QTextEdit").triggered.connect(this, QtEvents.newQTextEdit);
            sub3.addAction("QWebView (CNN)").triggered.connect(this, QtEvents.newQWebViewCNN);
            sub3.addAction("QWebView (Lively)").triggered.connect(this, QtEvents.newQWebViewLively);
            sub3.addSeparator();
            sub3.addAction("Any QWidget...").triggered.connect(this, QtEvents.newAnyWidget);
        var sub4 = menu.addMenu("Applications and Mashups");
            sub4.addAction("Asteroids").triggered.connect(this, QtEvents.newAsteroids);
            sub4.addAction("BBC Weather").triggered.connect(this, QtEvents.newWeather);
            sub4.addAction("Digital Clock").triggered.connect(this, QtEvents.newClock);
            sub4.addAction("Flickr").triggered.connect(this, QtEvents.newFlickr);
            sub4.addAction("Comics").triggered.connect(this, QtEvents.newComics);
            sub4.addAction("Finland Weather Cameras").triggered.connect(this, QtEvents.newWeatherCameras);
            sub4.addAction("Google Maps").triggered.connect(this, QtEvents.newMap);
            sub4.addAction("Map News").triggered.connect(this, QtEvents.newMapNews);
            sub4.addAction("Newsreader").triggered.connect(this, QtEvents.newRSS);
            sub4.addAction("Scrapbook").triggered.connect(this, QtEvents.newScrapBook);
            sub4.addAction("Web Camera Maps").triggered.connect(this, QtEvents.newWebcamMap);
        menu.addAction("New Subworld").triggered.connect(this, QtEvents.newSubworld);
        menu.addSeparator();
        var sub5 = menu.addMenu("Background Style")
            sub5.addAction("Fill Color...").triggered.connect(this, QtEvents.changeFillColor);
            sub5.addAction("Border Color...").triggered.connect(this, QtEvents.changeBorderColor);
            var sub51 = sub5.addMenu("Border Width");
                sub51.addAction("0").triggered.connect(this, QtEvents.changeBorderWidthTo0);
                sub51.addAction("1").triggered.connect(this, QtEvents.changeBorderWidthTo1);
                sub51.addAction("2").triggered.connect(this, QtEvents.changeBorderWidthTo2);
                sub51.addAction("3").triggered.connect(this, QtEvents.changeBorderWidthTo3);
                sub51.addAction("4").triggered.connect(this, QtEvents.changeBorderWidthTo4);
                sub51.addAction("5").triggered.connect(this, QtEvents.changeBorderWidthTo5);
                sub51.addAction("6").triggered.connect(this, QtEvents.changeBorderWidthTo6);
            var sub52 = sub5.addMenu("Opacity");
                sub52.addAction("5%").triggered.connect(this, QtEvents.changeOpacityTo5);
                sub52.addAction("20%").triggered.connect(this, QtEvents.changeOpacityTo20);
                sub52.addAction("40%").triggered.connect(this, QtEvents.changeOpacityTo40);
                sub52.addAction("60%").triggered.connect(this, QtEvents.changeOpacityTo60);
                sub52.addAction("80%").triggered.connect(this, QtEvents.changeOpacityTo80);
                sub52.addAction("100%").triggered.connect(this, QtEvents.changeOpacityTo100);
        var sub6 = menu.addMenu("Advanced")
            sub6.addAction("Load Script File...").triggered.connect(this, QtEvents.loadScriptFile);
            sub6.addSeparator();

            if (QtWorldMorph.timer)
                 sub6.addAction("Stop Timers").triggered.connect(this, QtEvents.toggleTimers);
            else sub6.addAction("Start Timers").triggered.connect(this, QtEvents.toggleTimers);
            
            if (QtEvents.wormholesEnabled)
                 sub6.addAction("Disable Wormholes").triggered.connect(this, QtEvents.toggleWormholes);
            else sub6.addAction("Enable Wormholes").triggered.connect(this, QtEvents.toggleWormholes);

            if (QtWorldMorph.wheelZoomingEnabled)
                 sub6.addAction("Disable Wheel Zooming").triggered.connect(this, QtEvents.toggleWheelZooming);
            else sub6.addAction("Enable Wheel Zooming").triggered.connect(this, QtEvents.toggleWheelZooming);

            sub6.addSeparator();
            sub6.addAction("Remove All Objects").triggered.connect(this, QtEvents.removeAllObjects);
            sub6.addAction("Unlively").triggered.connect(this, QtEvents.goUnlively);
        menu.addSeparator();
        menu.addAction("Exit System").triggered.connect(this, QtEvents.exitSystem);
        menu.popup(QCursor.pos());
    }

});

/**
  * QtWorldMorph subworld creation operations
  */

QtWorldMorph.addMethods({

    // The following function adds a subworld to an existing world.
    // It creates a button pointing to the subworld, and also places
    // a similar button in the subworld for navigating back to the
    // "upper" world.  Buttons are placed in the same place in both worlds.

    // Parameter buttonPos: qpt indicating the position of the buttons
    // Parameter buttonSize: qrect indicating the size of the buttons
    // Parameter buttonText: Text used in the push button pointing to subworld
    // Parameter backButtonText: Text used in the back button in the subworld
    // NOTE: All the parameters above are mandatory
    addSubworld: function(buttonPos, buttonSize, buttonText, backButtonText) {

        if (!buttonPos || !buttonSize || !buttonText || !backButtonText) return null;

        var subworld = new QtWorldMorph(); // Use default parameters
        var button = new QtProxyMorph(new QPushButton(buttonText), buttonSize);
        var backButton = new QtProxyMorph(new QPushButton(backButtonText), buttonSize);

        this.addMorph(button);
        button.setPos(buttonPos);
        
        subworld.addMorph(backButton);
        backButton.moveTo(10, 10); // setPos(buttonPos);

        // Connect the buttons to the worlds
        button.inner.pressed.connect(subworld, subworld.setCurrent);
        backButton.inner.pressed.connect(this, this.setCurrent);

        // TODO: Subworlds cannot be copied yet
        button.allowsCopying = false;
        backButton.allowsCopying = false;

        // Enable "wormhole" feature for the buttons
        button.wormhole = subworld;
        backButton.wormhole = this;

        // Keep track of subworlds associated with this world
        // TODO: Make it possible to remove / clean up subworlds!
        this.subworlds.push(subworld);

        return subworld;
    }

});

/*======================================================================
 * Class QtEvaluatorMorph (the code evaluator)
 * Allows JavaScript code to be edited and evaluated interactively
 *====================================================================*/

QtRectangleMorph.subclass("QtEvaluatorMorph", {

    // By default, leave 5 pixels for border
    inset: 5,    

    // Parameter rect: a QRect object (defines the bounds of the rectangle)
    // Parameter fillColor: a QColor object
    // Parameter borderWidth: a Number
    // Parameter borderColor: a QColor object
    // If any of the parameters is null/undefined, we use default values
    initialize: function($super, rect, fillColor, borderWidth, borderColor) {
        $super(rect, fillColor, borderWidth, borderColor);
        
        var bounds = this.rect();
        var inset = this.inset;
        
        this.textEditor = new QtProxyMorph(new QTextEdit(), 
             qrect(0, 0, bounds.width()-inset*2, bounds.height() * 0.75 - inset),   0);
        this.pushButton = new QtProxyMorph(new QPushButton("Evaluate"),
             qrect(0, 0, bounds.width()-inset*2, bounds.height() * 0.25 - inset*2), 0);

        this.textEditor.moveTo(inset, inset);
        this.pushButton.moveTo(inset, inset + bounds.height() * 0.75);

        this.textEditor.setBorderWidth(0);
        this.pushButton.setBorderWidth(0);

        this.textEditor.inner.plainText = "/* Enter JavaScript code here.\nPress the button below to evaluate.\nUse 'that' pointer to refer to the most recently manipulated object. */\n\nthat.rotate(10);"; 
        this.pushButton.inner.pressed.connect(this, this.evaluateCode);

        this.addMorph(this.textEditor);
        this.addMorph(this.pushButton);
        
        return this;
    },
    
    evaluateCode: function(foo) {
        var code = this.textEditor.inner.plainText;
        if (code) {
            try {
                eval(code);
            } catch(e) {
                // Insert possible error messages into the text editor itself
                this.textEditor.inner.insertPlainText(e);
            }
        }
    },
    
    // Change the bounds (size) of the code evaluator
    // The widgets inside are resized accordingly
    // TODO: We should use Qt's layouts here!
    // Parameter: QRectF
    setRect: function($super, rect) {
        $super(rect);

        var inset = this.inset;
        this.textEditor.setRect(qrect(0, 0, rect.width()-inset*2, rect.height() * 0.75 - inset),   0);
        this.pushButton.setRect(qrect(0, 0, rect.width()-inset*2, rect.height() * 0.25 - inset*2), 0);

        this.textEditor.moveTo(inset, inset);
        this.pushButton.moveTo(inset, inset + rect.height() * 0.75);

        return this;
    },
    
    // Change the border size (inset) of the code evaluator
    setInset: function(newInset) {
        this.inset = (newInset > 0) ? newInset : 0;
        this.setRect(this.rect());
        return this;
    },

    copy: function() {
        if (!this.peer) return null;
        var other = new QtEvaluatorMorph(this.rect(), this.getFillColor(),
                        this.getBorderWidth(), this.getBorderColor());

        // Opacity must be copied explicitly
        other.setOpacity(this.getOpacity());

        // Copy location and transformation info
        this.copyTransformationsTo(other);

        // Copy interaction properties
        this.copyInteractionPropertiesTo(other);

        // this.copyChildrenTo(other); // TODO: Does not work, because the
        // constructor above already creates the necessary children.
        // This would duplicate the submorphs! 
        return other;
    }

});

/*======================================================================
 * Class QtEvents
 * A singleton class that defines callback functions for various
 * Qt event handling functions (mostly for popup menus).
 * 
 * The functions are placed here because the original event handlers 
 * operate in the context of a morph's peer (QGraphicsItem).  
 * Consequently, we cannot call these functions via the 'this' pointer
 * as the programmer might expect.  Rather, we must have a globally 
 * accessible place for them. 
 *====================================================================*/

// NOTE: This class is a singleton -- do not instantiate!
Object.subclass("QtEvents", {
    documentation: "Mapping to native event system"
});

Object.extend(QtEvents, {

    // Locks all the Lively event handling (e.g., for shipping applications)
    userInterfaceLocked: false,

    // Enables drag-and-drop object transport between worlds
    wormholesEnabled: true,

    // Holds the morph that received the most recent mouse event.
    // This value is the same as in 'that' pointer, except that
    // 'currentMorph' has a value only when mouse button is down.
    currentMorph: null,

    // The amount of pixels in the corners of objects reserved 
    // for resizing, rotation and scaling (8x8); this is also
    // the minimum allowed size of an object (8x8 pixels)
    precision: 8,

    // This variable holds the position of the latest mouse down
    // (or context menu) event.  It is used, e.g., to avoid difficulties 
    // in passing parameters from the actual event handler callbacks
    // to the functions below.
    lastMouseDownPos: qpt(0, 0),
    
    // Difference between last mouse press position and the
    // position of the object inside which the press occurred 
    eventPosDiff: qpt(0, 0),

    // This variable is used for distinguishing between left and right
    // mouse button clicks.  (Qt calls regular mouse handers even when
    // the click would be processed also by a contextMenuEvent handler.)
    leftMouseButtonPressed: false,
    
    // Indicates whether the mouse was dragged, i.e., moved while
    // the left mouse button was held down
    mouseWasDragged: false,

    // Indicates whether an interactive resize/rotate/scale operation
    // is currently in progress
    transformationInProgress: false,

        // Values for the field above
        RESIZING: 1,
        ROTATING: 2,
        SCALING:  3,
        SHEARING: 4,

    // Help text that is shown near the mouse cursor while performing
    // interactive transformations (resizing, rotation, scaling, ...)
    helpTextItem: null,

    // VERY IMPORTANT: The event handler callbacks below are executed
    // in the context of the peer, i.e., 'this' pointer points 
    // to the peer (QGraphicsItem) rather than to the QtMorph!
    // The morph can be accessed via PeerToMorphTable or via the
    // QtEvents.currentMorph variable defined above. 

    unimplemented: function() {
        var box = new QMessageBox(GlobalWindow);
        box.text = "Unimplemented Feature";
        box.icon = QMessageBox.Information;
        box.exec();
        print("Invoked an unimplemented feature!");
    },

    noop: function() {
        // Do nothing!
    },

    /*
     * Operations invoked from QtWorldMorph popup menus
     */     

    newRectangle: function() {
        var newObject = new QtRectangleMorph(); // Use default parameters
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newEllipse: function() {
        var newObject = new QtEllipseMorph(); // Use default parameters
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newEditor: function() {
        var newObject = new QtEvaluatorMorph(qrect(0, 0, 250, 150));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newInspector: function() {
        try {
            if (QtInspector === undefined) return;
        } catch(e) {
            return; // Avoid errors if the inspector isn't available
        }

        var qWidget = new QtInspector();
        var newObject = new QtProxyMorph(qWidget);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newClassBrowser: function() {
        try {
            if (QtClassBrowser === undefined) return;
        } catch(e) {
            return; // Avoid errors if the browser isn't available
        }

        var qWidget = new QtClassBrowser();
        var newObject = new QtProxyMorph(qWidget);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    openDebugger: function() {
        debugger;
    },

    newSubworld: function() {
        var world = QtWorldMorph.current();
        if (world) world.addSubworld(QtEvents.lastMouseDownPos, qrect(0, 0, 160, 40), "Enter Subworld", "Back to Other World");
        that = null; // Disable 'that' (would not do the expected thing)
    },

    newClock: function() {
        var newObject = new QtClockMorph();
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newAsteroids: function() {
        // TODO: This demo is not re-entrant yet.
        // You can only create one instance of the game!

        // If the demo is not loaded, return without error messages
        try {
            if (Asteroids === undefined) return;
        } catch(e) {
            return;
        }

        asteroidGame = new Asteroids();
        var newObject = new QtProxyMorph(asteroidGame);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newWeather: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (WeatherWidget === undefined) return;
        } catch(e) {
            return;
        }

        var qweather = new WeatherWidget();
        var newObject = new QtProxyMorph(qweather, qrect(0, 0, 300, 420));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newWeatherCameras: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (QWeatherCameras === undefined) return;
        } catch(e) {
            return;
        }

        var qweather = new QWeatherCameras();
        var newObject = new QtProxyMorph(qweather, qrect(0, 0, 300, 420));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newScrapBook: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (ScrapBook === undefined) return;
        } catch(e) {
            return;
        }

        var qf = new ScrapBook();
        var newObject = new QtProxyMorph(qf, qrect(0, 0, 400, 420));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newWebcamMap: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (QWebcamMap === undefined) return;
        } catch(e) {
            return;
        }

        var qf = new QWebcamMap();
        var newObject = new QtProxyMorph(qf, qrect(0, 0, 400, 420));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newFlickr: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (FlickrWidget === undefined) return;
        } catch(e) {
            return;
        }

        var qf = new FlickrWidget();
        var newObject = new QtProxyMorph(qf, qrect(0, 0, 400, 420));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newComics: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (ComicWidget === undefined) return;
        } catch(e) {
            return;
        }

        var qcomic = new ComicWidget();
        var newObject = new QtProxyMorph(qcomic, qrect(0, 0, 400, 420));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newRSS: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (RSSWidget === undefined) return;
        } catch(e) {
            return;
        }

        var qrss = new RSSWidget();
        var newObject = new QtProxyMorph(qrss);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newMap: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (QMap === undefined) return;
        } catch(e) {
            return;
        }

        var qmap = new QMap();
        var newObject = new QtProxyMorph(qmap);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newMapNews: function() {
        // If the demo is not loaded, return without error messages
        try {
            if (QMapNews === undefined) return;
        } catch(e) {
            return;
        }

        var qmap = new QMapNews();
        var newObject = new QtProxyMorph(qmap);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newQCalendarWidget: function() {
        var qWidget = new QCalendarWidget();
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 250, 200));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newQCheckBox: function() {
        var qWidget = new QCheckBox("Toggle Me!");
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 90, 30));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQColorDialog: function() {
        var qWidget = new QColorDialog();
        qWidget.setOption(QColorDialog.ShowAlphaChannel, true);
        qWidget.setOption(QColorDialog.NoButtons, true);
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 510, 356));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },
    
    newQComboBox: function() {
        var qWidget = new QComboBox();
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 100, 30));
        newObject.inner.addItem("Item 1"); // 'inner' points to qWidget
        newObject.inner.addItem("Item 2");
        newObject.inner.addItem("Item 3");
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQDateTimeEdit: function() {
        var qWidget = new QDateTimeEdit();
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 150, 30));
        newObject.inner.date = new Date();
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQDial: function() {
        var qWidget = new QDial();
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 120, 120));
        newObject.inner.notchesVisible = true;
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQLabel: function() {
        var pixmap = new QPixmap("qt-logo.png"); // "weather.png");
        if (!pixmap) return; 
        var qWidget = new QLabel();
        qWidget.setPixmap(pixmap);
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 80+10, 80+10)); // WAS: 160+10, 110+10
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQLineEdit: function() {
        var qWidget = new QLineEdit("Edit Me!");
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 170, 30));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQProgressBar: function() {
        var qWidget = new QProgressBar();
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 220, 30));
        newObject.inner.setValue(50);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQPushButton: function() {
        var qWidget = new QPushButton("Push Me!");
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 110, 35));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQScrollBar: function() {
        var qWidget = new QScrollBar(Qt.Vertical);
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 30, 110));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQSlider: function() {
        var qWidget = new QSlider(Qt.Vertical);
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 40, 110));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQSpinBox: function() {
        var qWidget = new QSpinBox();
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 60, 40));
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQTextEdit: function() {
        var qWidget = new QTextEdit("Edit Me!");
        var newObject = new QtProxyMorph(qWidget, qrect(0, 0, 200, 200), 8);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);    
    },

    newQWebViewCNN: function() {
        var qWidget = new QWebView();
        qWidget.resize(600, 400);
        var newObject = new QtProxyMorph(qWidget, null, 8);
        newObject.setPos(QtEvents.lastMouseDownPos);
        newObject.inner.load(new QUrl("http://www.cnn.com/"));
        that = QtWorldMorph.current().addMorph(newObject);
    },

    newQWebViewLively: function() {
        var qWidget = new QWebView();
        qWidget.resize(600, 400);
        var newObject = new QtProxyMorph(qWidget, null, 8);
        newObject.setPos(QtEvents.lastMouseDownPos);
        newObject.inner.load(new QUrl("http://research.sun.com/projects/lively"));
        that = QtWorldMorph.current().addMorph(newObject);
    },
    
    newAnyWidget: function() {
        QtEvents.dialog = new QDialog(GlobalWindow);
        QtEvents.dialog.setWindowTitle("Enter QWidget Class Name");
        var layout = new QGridLayout(QtEvents.dialog);
        QtEvents.editor = new QLineEdit("QCalendarWidget", QtEvents.dialog);
        QtEvents.editor.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum);
        var bDone = new QPushButton("Create", QtEvents.dialog);
        layout.addWidget(QtEvents.editor, 0, 0);
        layout.addWidget(bDone, 0, 1);
        QtEvents.dialog.setLayout(layout);
        bDone.pressed.connect(QtEvents, QtEvents.createAnyQWidget);
        QtEvents.dialog.exec();
    },
    
    createAnyQWidget: function() {
        // Prevent creation of multiple widgets
        if (QtEvents.dialog) {
            var dialog = QtEvents.dialog;
            QtEvents.dialog = null;
            dialog.accept();
        } else return;

        var className = QtEvents.editor.text;
        try {
            className = eval(className);
            if (className === undefined) return;
        } catch(e) {
            return;
        }
        
        var qf = new className();
        var rect = qf.rect;
        var newObject = new QtProxyMorph(qf, rect);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },
    
    loadScriptFile: function() {
        QtEvents.dialog = new QDialog(GlobalWindow);
        QtEvents.dialog.setWindowTitle("Load Script File");
        var layout = new QGridLayout(QtEvents.dialog);
        QtEvents.editor = new QLineEdit("Enter filename", QtEvents.dialog);
        QtEvents.editor.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum);
        var bDone = new QPushButton("Load", QtEvents.dialog);
        layout.addWidget(QtEvents.editor, 0, 0);
        layout.addWidget(bDone, 0, 1);
        QtEvents.dialog.setLayout(layout);
        var fileName = QFileDialog.getOpenFileName(GlobalWindow);
        if (fileName.length != 0 ) {
            QtEvents.editor.text = fileName;
        }
        bDone.pressed.connect(QtEvents, QtEvents.evalScriptFile);
        QtEvents.dialog.exec();
    },
    
    evalScriptFile: function() {
        // Prevent multiple invocations
        if (QtEvents.dialog) {
            var dialog = QtEvents.dialog;
            QtEvents.dialog = null;
            dialog.accept();
        } else return;

        var filename = QtEvents.editor.text;
        var file = new QFile(filename);
        
        if (!file.open(new QIODevice.OpenMode(QIODevice.ReadOnly, QIODevice.Text))) {
            message = "Cannot read file " + filename + ":\n" + file.errorString();
            QMessageBox.warning(GlobalWindow, "Application", message);
            return;
        }
        
        var inStream = new QTextStream(file);
        var text = inStream.readAll();
        file.close();
        
        try {
            if (text === undefined) return;
        } catch(e) {
            return;
        }
        
        var qf;
        try {
            var value = eval(text);
            if (value) {
                // is QWidget?
                if (value.inherits && value.inherits("QWidget")) {
                    qf = value;
                } else {
                    // creates the object immediately with the constructor
                    qf = new value();
                }
                // Keep the source code available (increases memory footprint...)
                qf.source = text; 
            }
        } catch (e) {
            return;
        }
        
        if (qf.inherits && qf.inherits("QWidget")) {
            var rect = qf.rect;
            var newObject = new QtProxyMorph(qf, rect);
            newObject.setPos(QtEvents.lastMouseDownPos);
            that = QtWorldMorph.current().addMorph(newObject);
        }
    },

    // TODO: Implement a real timer/scheduler mechanism later
    toggleTimers: function() {
        
        if (QtWorldMorph.timer) {
            // Stop existing timer(s)
            QtWorldMorph.timer.stop();
            QtWorldMorph.timer = null;
            
            try { // TODO: This only stops the most recent game instance
                asteroidGame.timer.stop();
            } catch(e) {}

        } else {
            // Start timer(s)
            QtWorldMorph.timer = new QTimer(GlobalWindow);
            QtWorldMorph.timer.timeout.connect(this, doStepping);
            QtWorldMorph.timer.start(20);

            try { // TODO: This only starts the most recent game instance
                asteroidGame.timer.start(50);
            } catch(e) {}

        }

    },

    removeAllObjects: function() {
        var box = new QMessageBox(GlobalWindow);
        box.text = "Are you sure you want to remove all the objects?";
        box.standardButtons = QMessageBox.Ok | QMessageBox.Cancel;
        box.icon = QMessageBox.Question;
        var result = box.exec();

        if (result == QMessageBox.Ok) {
            QtWorldMorph.current().removeAllMorphs();
        }
    },

    exitSystem: function() {
        var box = new QMessageBox(GlobalWindow);
        box.text = "Are you sure you want to exit the system?";
        box.standardButtons = QMessageBox.Ok | QMessageBox.Cancel;
        box.icon = QMessageBox.Question;
        var result = box.exec();

        if (result == QMessageBox.Ok) {
            QCoreApplication.exit(0);
        }
    },

    /*
     * Operations invoked from QtMorph popup menus
     */     

    inspectThisObject: function() {
        try {
            if (QtInspector === undefined) return;
        } catch(e) {
            return; // Avoid errors if the inspector isn't available
        }

        var thisMorph = QtEvents.currentMorph;
        if (!thisMorph) return;

        // Open inspector in the context of this morph
        // TODO: The inspector does not quite work with this feature yet
        var qWidget = new QtInspector(null, thisMorph, thisMorph.toString());

        var newObject = new QtProxyMorph(qWidget);
        newObject.setPos(QtEvents.lastMouseDownPos);
        that = QtWorldMorph.current().addMorph(newObject);
    },

    shrinkByTwo: function() {
        that = QtEvents.currentMorph;
        if (that.allowsScaling) this.scale(0.5, 0.5);
    },

    enlargeByTwo: function() {
        that = QtEvents.currentMorph;
        if (that.allowsScaling) this.scale(2, 2);
    },

    rotateLeft: function() {
        that = QtEvents.currentMorph;
        if (that.allowsRotation) this.rotate(-10);
    },

    rotateRight: function() {
        that = QtEvents.currentMorph;
        if (that.allowsRotation) this.rotate(10);
    },

    shearLeft: function() {
        that = QtEvents.currentMorph;
        if (that.allowsShearing) this.shear(-0.1, 0);
    },

    shearRight: function() {
        that = QtEvents.currentMorph;
        if (that.allowsShearing) this.shear(0.1, 0);
    },

    resetTransformations: function() {
        that = QtEvents.currentMorph;
        this.resetTransform();
    },

    changeFillColor: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            var fillColor = currentMorph.getFillColor();
            var color = QColorDialog.getColor(fillColor, GlobalWindow, "Change Fill Color", QColorDialog.ShowAlphaChannel);
            if (color.isValid()) currentMorph.setFillColor(color);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderColor: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            var fillColor = currentMorph.getBorderColor();
            var color = QColorDialog.getColor(fillColor, GlobalWindow, "Change Fill Color", QColorDialog.ShowAlphaChannel);
            if (color.isValid()) currentMorph.setBorderColor(color);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderWidthTo0: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setBorderWidth(0);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderWidthTo1: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setBorderWidth(1);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderWidthTo2: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setBorderWidth(2);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderWidthTo3: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setBorderWidth(3);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderWidthTo4: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setBorderWidth(4);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderWidthTo5: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setBorderWidth(5);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeBorderWidthTo6: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setBorderWidth(6);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeOpacityTo5: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setOpacity(0.05); // Don't want to make the object disappear entirely!
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeOpacityTo20: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setOpacity(0.2);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeOpacityTo40: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setOpacity(0.4);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeOpacityTo60: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setOpacity(0.6);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeOpacityTo80: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setOpacity(0.8);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeOpacityTo100: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setOpacity(1.0);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeInsetTo0: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setInset(0);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeInsetTo2: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setInset(2);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeInsetTo4: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setInset(4);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeInsetTo6: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setInset(6);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeInsetTo8: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setInset(8);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    changeInsetTo10: function() {
        var currentMorph = QtEvents.currentMorph;
        if (currentMorph) {
            currentMorph.setInset(10);
        }
        if (currentMorph !== QtWorldMorph.current()) that = currentMorph;
    },

    copyMorphToWorld: function() {
        // Get the item's backpointer to the morph
        var thisMorph = QtEvents.currentMorph;
        if (!thisMorph || !thisMorph.allowsCopying) return;

        // Copy and add to world (at cursor location)
        var newMorph = thisMorph.copy();
        if (newMorph) { 
            QtWorldMorph.current().addMorph(newMorph);
            newMorph.setPos(QtEvents.lastMouseDownPos);
            that = newMorph;
        }
    },

    removeMorphFromWorld: function() {
        var thisMorph = QtEvents.currentMorph;
        if (thisMorph) thisMorph.remove();
    },

    whoAmI: function() {
        var thisMorph = QtEvents.currentMorph;
        var description = thisMorph.toString();
        if (description) this.setToolTip(description)
        print(this + " in " + QtEvents.currentMorph);
        that = QtEvents.currentMorph;
    },

    showPropertiesDialog: function() {

        // Make sure 'that' points to currentMorph
        that = QtEvents.currentMorph;
        if (!that) return;

        var dialog = new QDialog(GlobalWindow);
        dialog.setWindowTitle("Interaction Properties");

        var c1 = new QCheckBox("Can be moved", dialog);
        var c2 = new QCheckBox("Can be copied", dialog);
        var c3 = new QCheckBox("Can be resized", dialog);
        var c4 = new QCheckBox("Can be rotated", dialog);
        var c5 = new QCheckBox("Can be scaled", dialog);
        var c6 = new QCheckBox("Can be sheared", dialog);
        var c7 = new QCheckBox("Can be attached to other objects", dialog);
        var c8 = new QCheckBox("Other objects can be attached to me", dialog);
        var bDone = new QPushButton("Done", dialog);

        bDone.pressed.connect(dialog, dialog.accept);
        // bCancel.pressed.connect(dialog, dialog.reject);

        // Read the property values from the actual object
        c1.setChecked(that.allowsMoving);
        c2.setChecked(that.allowsCopying);
        c3.setChecked(that.allowsResizing); 
        c4.setChecked(that.allowsRotation);
        c5.setChecked(that.allowsScaling);
        c6.setChecked(that.allowsShearing);
        c7.setChecked(that.allowsAttachingToOthers);
        c8.setChecked(that.allowsAttachingOfOthers);

        // Connect the check boxes changes to correct value changes 
        c1.stateChanged.connect(that, that.setAllowsMoving);
        c2.stateChanged.connect(that, that.setAllowsCopying);
        c3.stateChanged.connect(that, that.setAllowsResizing);
        c4.stateChanged.connect(that, that.setAllowsRotation);
        c5.stateChanged.connect(that, that.setAllowsScaling);
        c6.stateChanged.connect(that, that.setAllowsShearing);
        c7.stateChanged.connect(that, that.setAllowsAttachingToOthers);
        c8.stateChanged.connect(that, that.setAllowsAttachingOfOthers);

        var layout = new QVBoxLayout();
        layout.sizeConstraint = QLayout.SetFixedSize;
        layout.addWidget(c1, 0, 0);
        layout.addWidget(c2, 0, 0);
        layout.addWidget(c3, 0, 0);
        layout.addWidget(c4, 0, 0);
        layout.addWidget(c5, 0, 0);
        layout.addWidget(c6, 0, 0);
        layout.addWidget(c7, 0, 0);
        layout.addWidget(c8, 0, 0);
        layout.addWidget(bDone, 0, 0);
        dialog.setLayout(layout);

        dialog.exec();
    },

    lockUI: function(state) {
        if (state == Qt.Unchecked) QtEvents.userInterfaceLocked = false;
        else if (state == Qt.Checked) QtEvents.userInterfaceLocked = true;
    },

    // Function that disables all the lively functionality
    // All the underlying application (e.g., QWidget) behavior
    // remains fully active.
    goUnlively: function() {
        var box = new QMessageBox(GlobalWindow);
        box.text = "Warning: You are about to remove all the Lively capabilities from the system.";
        box.informativeText = "Are you sure you want to continue?";
        box.standardButtons = QMessageBox.Ok | QMessageBox.Cancel;
        box.icon = QMessageBox.Question;
        var result = box.exec();

        if (result == QMessageBox.Ok) {
            QtEvents.userInterfaceLocked = true;
        }
    },

/* Not used currently
    setWheelZooming: function(state) {
        if (state == Qt.Unchecked) QtWorldMorph.wheelZoomingEnabled = false;
        else if (state == Qt.Checked) QtWorldMorph.wheelZoomingEnabled = true;
    },
    
    setWormholes: function(state) {
        if (state == Qt.Unchecked) QtEvents.wormholesEnabled = false;
        else if (state == Qt.Checked) QtEvents.wormholesEnabled = true;
    },
*/

    // Toggle world zooming with the mouse wheel on/off
    toggleWheelZooming: function() {
        if (QtWorldMorph.wheelZoomingEnabled)
             QtWorldMorph.wheelZoomingEnabled = false;
        else QtWorldMorph.wheelZoomingEnabled = true;
    },

    // Toggle the wormhole feature (transporting objects across worlds) on/off
    toggleWormholes: function() {
        if (QtEvents.wormholesEnabled)
             QtEvents.wormholesEnabled = false;
        else QtEvents.wormholesEnabled = true;    
    },
    
    // This function is used for opening all the mashup demos easily
    openMashupDemos: function() {
        QtEvents.lastMouseDownPos = qpt(50, 50);
        QtEvents.newFlickr();
        QtEvents.lastMouseDownPos = qpt(60, 60);
        QtEvents.newComics();
        QtEvents.lastMouseDownPos = qpt(70, 70);
        QtEvents.newWeatherCameras();
        QtEvents.lastMouseDownPos = qpt(80, 80);
        QtEvents.newWeather();
        QtEvents.lastMouseDownPos = qpt(90, 90);
        QtEvents.newMap();
        QtEvents.lastMouseDownPos = qpt(100, 100);
        QtEvents.newRSS();
        QtEvents.lastMouseDownPos = qpt(110, 110);
        QtEvents.newMapNews();
        QtEvents.lastMouseDownPos = qpt(120, 120);
        QtEvents.newScrapBook();
    }

});

/*======================================================================
 * Class QtClockMorph
 * A Clock Morph
 *====================================================================*/

QtProxyMorph.subclass("QtClockMorph", {

    documentation: "a QtClockMorph",
    
    initialize: function($super) {

        // Create clock face
        var display = new QLCDNumber(8); 
        display.segmentStyle = QLCDNumber.Flat;
        $super(display, qrect(0, 0, 120, 40)); 
        this.updateTime();

        // Create timer to run the clock
        this.timer = new QTimer(GlobalWindow);
        this.timer.timeout.connect(this, this.updateTime);
        this.timer.start(1000);

        return this;
    },

    // Update clock face
    updateTime: function() {
        var time = QTime.currentTime();
        var format = "hh:mm";
        format += ((time.second() % 2) == 0) ? ":" : " ";
        format += "ss";
        var text = time.toString(format);
        this.inner.display(text);
    }

});

/*======================================================================
 * Main Window Creation
 *====================================================================*/

// The code below creates and initializes the
// main window that will contain the world(s)

function MainWindow() {
    QWidget.call(this);

    // Set window (QWidget) attributes    
    this.setWindowTitle("Lively for Qt (May 11, 2009)");
    this.resize(QtWorldMorph.defaultExtent.x, QtWorldMorph.defaultExtent.y);
    this.mouseTracking = true;

    // Define the window resize handler
    this.resizeEvent = resizeMainWindow;

    // Show the (native) window
    this.show();

    // Bind event handler callback for keyPressEvent
    MainWindow.prototype.keyPressEvent = MainWindow_keyDownEvent;

    return this;
}

// MainWindow is a subclass of QWidget
MainWindow.prototype = new QWidget();

// Key event handler
// (used for toggling between normal and full screen mode
// on Nokia Maemo devices)
MainWindow_keyDownEvent = function(event) {
    var key = event.key();

    if (key == Qt.Key_F6) {
        this.toggleWindowState();
    }
}

MainWindow.prototype.toggleWindowState = function() {
    var state = this.windowState() ^ Qt.WindowFullScreen;
    /* State is 4 when fullscreen, 0 in normal state */
    if (state != 4) {
        this.setWindowState(Qt.WindowNoState | !Qt.WindowFullScreen);
    } else {
        this.setWindowState(Qt.WindowFullScreen);
    }
}

// Adjust world size automatically when the main window is resized
// TODO: The size of the other worlds is not adjusted immediately
function resizeMainWindow(qtEvent) {
    if (!GlobalWindow) return;

    // Read new window size
    var size = qtEvent.size();
    var newWidth = size.width();
    var newHeight = size.height();

    // Adjust view, scene and world size accordingly
    var world = QtWorldMorph.current();
    var view = world.view;
    var scene = world.scene;

    scene.setSceneRect(2, 2, newWidth-4, newHeight-4); // Avoid scroll bars!
    world.peer.setRect(2, 2, newWidth-4, newHeight-4);
    view.resize(newWidth, newHeight);
}

/*==============================================================================
 * Genesis
 *============================================================================*/

// Create the main window with its view and scene
GlobalWindow = new MainWindow();

// Create a gradient color for the world background
var wgrad = new QLinearGradient(0, 0, 0, QtWorldMorph.defaultExtent.y);
wgrad.setColorAt(0, new QColor("white"));
wgrad.setColorAt(1, new QColor(0x53, 0x82, 0xA1));
wgrad.setSpread(QGradient.ReflectSpread);

// Create the world and display it on the scene
var world = new QtWorldMorph(null, wgrad);
world.setCurrent();

/*==============================================================================
 * Sample Objects and Applications
 *============================================================================*/

// Create sample QtMorphs
var clock = new QtClockMorph();
clock.moveTo(25, 40);
world.addMorph(clock);

var rect0 = new QtRectangleMorph(qrect(0, 0, 100, 30));
rect0.setPos(qpt(200, 40));
world.addMorph(rect0);

// IMPORTANT: Set the initial value of the 'that' pointer!
that = rect0;

var rect1 = new QtRectangleMorph(qrect(0, 0, 100, 30));
rect1.setPos(qpt(230, 80));
world.addMorph(rect1);

var rect2 = new QtRectangleMorph(qrect(0, 0, 100, 30));
rect2.setPos(qpt(180, 140));
world.addMorph(rect2);

var rect3 = new QtRectangleMorph(qrect(-50, -15, 100, 30), QtColor.green, 2, QtColor.black);
rect3.setPos(qpt(-25, -10)); // WAS: 400, 220
rect3.setOpacity(0.75);
// world.addMorph(rect3);

var rect4 = rect2.copy();
rect2.addMorph(rect4);
rect4.setPos(qpt(15, 15)).rotate(5).setFillColor(new QColor("indigo"));
// rect2.setFillColor(QtColor.darkCyan);

var rect5 = rect4.copy();
rect4.addMorph(rect5);
rect5.setPos(qpt(15, 15));

// Gradient color for the rectangle above
var gradient4 = new QLinearGradient(0, 0, 100, 0);
gradient4.setColorAt(0, new QColor(Qt.blue));
gradient4.setColorAt(1, new QColor(Qt.red));
gradient4.setSpread(QGradient.ReflectSpread);
rect5.setFillColor(gradient4);

var ellipse1 = new QtEllipseMorph(qrect(-50, -20, 100, 40), QtColor.transparent, 3, QtColor.blue);
ellipse1.setPos(qpt(85, 150));
world.addMorph(ellipse1);

var ellipse2 = new QtEllipseMorph(qrect(0, 0, 40, 40), QtColor.red, 2, QtColor.blue);
ellipse2.setOpacity(0.9);
ellipse2.setPos(qpt(0, 0));
ellipse1.addMorph(ellipse2);
ellipse1.addMorph(rect3);

var calendar = new QtProxyMorph(new QCalendarWidget(), qrect(0, 0, 250, 200), 6);
calendar.moveTo(360, 180);
calendar.shear(0.2, 0);
world.addMorph(calendar);

var slider = new QtProxyMorph(new QSlider(/*Qt.Horizontal*/), qrect(0, 0, 25+10, 100+10), 5);
slider.setPos(qpt(400, 40));
world.addMorph(slider);

var spinbox = new QtProxyMorph(new QSpinBox(), qrect(0, 0, 60+10, 30+10), 5);
spinbox.setPos(qpt(700, 100));
world.addMorph(spinbox);

var dial = new QtProxyMorph(new QDial(), qrect(0, 0, 100+10, 100+10), 5);
dial.inner.notchesVisible = true;
dial.setPos(qpt(550, 40));
world.addMorph(dial);

var progressbar = new QtProxyMorph(new QProgressBar(), qrect(0, 0, 200+10, 25+10), 5);
progressbar.setPos(qpt(700, 40));
world.addMorph(progressbar);

var scrollbar = new QtProxyMorph(new QScrollBar(/*Qt.Horizontal*/), qrect(0, 0, 25+10, 100+10), 5);
scrollbar.setPos(qpt(475, 40));
world.addMorph(scrollbar);

// Connect the widget value changes to each other
slider.inner.valueChanged.connect(this, spinbox.inner.setValue);
spinbox.inner['valueChanged(int)'].connect(this, dial.inner.setValue); // Qt 4.5 QSpinBox has two different 'valueChanged' methods!
dial.inner.valueChanged.connect(this, progressbar.inner.setValue);
progressbar.inner.valueChanged.connect(this, scrollbar.inner.setValue);
scrollbar.inner.valueChanged.connect(this, slider.inner.setValue);

// Set initial value for the widgets above
spinbox.inner.setValue(10);

var dateEdit = new QtProxyMorph(new QDateEdit(), qrect(0, 0, 100+10, 30+10), 5);
dateEdit.setPos(qpt(800, 100));
world.addMorph(dateEdit);

var evaluator = new QtEvaluatorMorph(qrect(0, 0, 250, 145));
evaluator.moveTo(10, 230);
world.addMorph(evaluator);

var qWebBrowser = new QWebView();
qWebBrowser.resize(600, 400);
var webBrowser = new QtProxyMorph(qWebBrowser, null, 8);
webBrowser.moveTo(220, 400);
webBrowser.rotate(5);
webBrowser.inner.load(new QUrl("http://www.cnn.com/"));
world.addMorph(webBrowser);

// Create a button that opens up all the mashup demos
var mashButton = new QtProxyMorph(new QPushButton("Open Mashup Demos"), qrect(0, 0, 170+10, 30+10), 5);
mashButton.moveTo(10, 425);
mashButton.setFillColor(QtColor.darkRed);
mashButton.inner.pressed.connect(this, QtEvents.openMashupDemos);
world.addMorph(mashButton);

// Define a subworld
var subworld = world.addSubworld(qpt(10, 475), qrect(0, 0, 160, 40), "Enter Subworld 1", "Back to Main World");

// Create a gradient color for the subworld background
var subgrad = new QLinearGradient(0, 0, 0, QtWorldMorph.defaultExtent.y);
subgrad.setColorAt(0, new QColor("white"));
subgrad.setColorAt(1, new QColor("cyan"));
subgrad.setSpread(QGradient.ReflectSpread);
subworld.setFillColor(subgrad);

// Define another subworld
var subworld2 = world.addSubworld(qpt(10, 525), qrect(0, 0, 160, 40), "Enter Subworld 2", "Back to Main World");
subgrad.setColorAt(1, new QColor("salmon")); // Modify gradient
subworld2.setFillColor(subgrad);

// Open an inspector also
try {
    if (QtInspector) {
        var inspector = new QtProxyMorph(new QtInspector());
        inspector.moveTo(700, 200);
        inspector.setRect(qrect(0, 0, 600, 500));
        world.addMorph(inspector);
    }
} catch(e) {
    // Avoid errors if the inspector isn't available
}

// Default timer behavior
function doStepping() {
    ellipse1.rotate(-6);
    rect3.rotate(10);
}

// Start timers automatically
QtEvents.toggleTimers();

/*==============================================================================
 * Qt startup
 *============================================================================*/

QCoreApplication.exec();

